/*******************************************************************************
  Copyright(c) 2000 - 2003 Radu Corlan. All rights reserved.
  
  This program is free software; you can redistribute it and/or modify it 
  under the terms of the GNU General Public License as published by the Free 
  Software Foundation; either version 2 of the License, or (at your option) 
  any later version.
  
  This program is distributed in the hope that it will be useful, but WITHOUT 
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
  more details.
  
  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc., 59 
  Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  
  The full GNU General Public License is included in this distribution in the
  file called LICENSE.
  
  Contact Information: radu@corlan.net
*******************************************************************************/

// batch.c: batch processing functions (photometry, frame ops)

// $Revision: 1.1 $
// $Date: 2003/12/01 00:19:37 $

#include <math.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <time.h>

#include "gcx.h"
#include "x11ops.h"

struct vs_recipy vsr;

#define MAX_VAR_REF 128

static struct multistar_stats {
	int n;	// nr of frames collected so far
	double sum[MAX_VAR_REF];
	double sumsq[MAX_VAR_REF];
	double errsum[MAX_VAR_REF];
	double errsumsq[MAX_VAR_REF];
	double maxerr[MAX_VAR_REF];
	double min[MAX_VAR_REF];
	double max[MAX_VAR_REF];
} mss;

// reset the stats structure
static void reset_star_stats(void)
{
	int i;
	mss.n = 0;
	for (i=0; i<MAX_VAR_REF; i++) {
		mss.sum[i] = 0.0;
		mss.sumsq[i] = 0.0;
		mss.errsum[i] = 0.0;
		mss.errsumsq[i] = 0.0;
		mss.maxerr[i] = -HUGE;
		mss.min[i] = HUGE;
		mss.max[i] = -HUGE;
	}
}


// collect stats on photometry results from frames
static void collect_star_stats(struct vs_recipy *vs)
{
	int i;
	double v;

	for (i=0; i < vs->cnt; i++) {
		if (vs->s[i].aph.flags & AP_STD_STAR) {
			v = vs->s[i].aph.residual;
		} else {
			v = vs->s[i].aph.stdmag;
		}
		mss.sum[i] += v;
		mss.sumsq[i] += sqr(v);
		mss.errsum[i] += vs->s[i].aph.magerr;
		mss.errsumsq[i] += sqr(vs->s[i].aph.magerr);
		if (mss.maxerr[i] < vs->s[i].aph.magerr)
			mss.maxerr[i] = vs->s[i].aph.magerr;
		if (mss.max[i] < v)
			mss.max[i] = v;
		if (mss.min[i] > v)
			mss.min[i] = v;
	}
	mss.n++;
}

// report the stats on a multiframe run
static void report_multistar_stats(FILE *fp, struct vs_recipy *vs)
{
	int i;
	double sm, n;


	n = 1.0 * mss.n;

	if (!fp)
		return;

	fprintf(fp, "\n# Multiframe run summary (%d frames)\n", mss.n);

	fprintf(fp, "# Star# avg_mag   sigma  min_res max_res avg_err max_err uncert\n");
	for (i=0; i<vs->cnt; i++) {
		if (vs->s[i].aph.flags & AP_STD_STAR) {
			sm = vs->s[i].aph.stdmag;
		} else {
			sm = 0;
		}
		fprintf(fp, "# %c%-4d %7.3f %7.3f %7.3f %7.3f %7.3f %7.3f %7.3f\n",
			(vs->s[i].aph.flags & AP_STD_STAR) ? 'S' : 'P',
			i,
			sm + mss.sum[i] / n,
			sqrt( mss.sumsq[i] / n - sqr(mss.sum[i] / n) ),
			mss.min[i] - mss.sum[i] / n,
			mss.max[i] - mss.sum[i] / n,
			mss.errsum[i] / n,
			mss.maxerr[i],
			sqrt( mss.sumsq[i] / n - sqr(mss.sum[i] / n) ) / sqrt(n)
			);
 	} 

}

// print the noise calculation summary for a run
static void print_noise_summary(FILE *fp, struct ccd_frame *fr, struct vs_recipy *vs)
{
	double t, d, am, sn;
	int i;

	if (!fp)
		return;

	get_scint_pars(fr, vs, &t, &d, &am);
	sn = scint_noise(fr, vs);

	fprintf(fp, "\n# Noise calculation summary for frame %s\n", fr->name);
	fprintf(fp, "# Scintillation noise is %.4f (from D=%.0f t=%.1f am=%.1f factor=%.1f)\n",
		sn, d, t, am, vs->scint_factor);
	fprintf(fp, "# Star#  stdmag  backgr   readn  photon   scint    std    total\n");
	for (i = 0; i < vs->cnt; i++) {
		fprintf(fp, "# %c%-4d %7.3f %7.4f %7.4f %7.4f %7.4f %7.4f %7.4f\n",
			(vs->s[i].aph.flags & AP_STD_STAR) ? 'S' : 'P',
			i,
			vs->s[i].aph.stdmag,
			vs->s[i].aph.sky_err * vs->s[i].aph.star_all / vs->s[i].aph.star,
			(fr->exp.rdnoise) * sqrt(1.0 * vs->s[i].aph.star_all) / vs->s[i].aph.star,
			sqrt(abs(vs->s[i].aph.tflux - fr->exp.bias 
				 * vs->s[i].aph.star_all) * fr->exp.scale) 
			/ fr->exp.scale / vs->s[i].aph.star,
			sn,
			(vs->s[i].aph.flags & AP_STD_STAR) ? 0 : vs->p.std_err,
			vs->s[i].aph.magerr
			);
	}
}

// search and center the recipy stars on the image
// return 0 if all stars were correctly matched, or the
// key used to exit the star search
static int find_recipy_stars(struct ccd_frame *fr, int iw, struct vs_recipy *vs, int fstd_indx)
{
	struct fr_display_par upar;
	int ret, nf;
	int noaut, i;
	double dx,dy;
	

	noaut = 0;
	memcpy(&upar, &iwt[iw].par, sizeof(struct fr_display_par) );
	upar.map_changed = 1;
	nf = get_star_near(fr, (int)vs->s[fstd_indx].x, 
			    (int)vs->s[fstd_indx].y, 
			    0, &upar.star);
	do {
		nf = 0;
		dx = upar.star.x - vs->s[fstd_indx].x;
		dy = upar.star.y - vs->s[fstd_indx].y;
		for (i=0; i<vs->cnt; i++) {
			vs->s[i].x += dx;
			vs->s[i].y += dy;
			ret = follow_star(fr, 8, &(vs->s[i]), &(vs->s[i]));
			if (ret) {
				info_printf("Star %d (%.1f %.1f) not found\n", i, 
					    vs->s[i].x, vs->s[i].y);
				nf = 1;
			}
		}
		ret = show_frame(iw, fr, &upar);
		release_frame(fr);
		draw_ap_stars(iw, vs, 0, 0);
	
		if (nf) {// we couldn;t find a star, stop and ask for it
			info_printf("At least one star not found; press 'p' to select new\n");
			info_printf("refernce star, 'c' to continue with current coords,\n");
			info_printf("'n' to reject frame or 'a' to resume auto mode\n");
			while ((ret = x_select_wait()) == 0);
		}
		memcpy(&upar, &iwt[iw].par, sizeof(struct fr_display_par) );
		upar.map_changed = 1;
	} while (nf && ret != 'c' && ret != 'a' && ret != 'n');
	return (nf ? ret : 0);
}

// aphot_files does aperture photometry on a series of files
// each frame is loaded for display; the position of the first standard star
// is shown on the display; the user is allowed to adjust the position if not satisfied
// then, all the standard and program, stars are shown, and the user is prompted for 
// confirmation. 
// results are reproted in repfn

int aphot_files(int ac, char **av, int ind, char *rf, char *of)
{
	int n=0;
	struct ccd_frame *fr = NULL;
	int ret, i;
	int autop = 0;
	int iw;
	int fstd_indx = 0;	// index of first standard star
	FILE *rfp, *ofp;
	struct vs_recipy *vs; 

	vs = &vsr;
	rfp = fopen(rf, "r");
	if (rfp) {
		fscan_recipy(rfp, vs);

	} else {
		err_printf("Cannot read recipy file: %s\n", rf);
		return ERR_FILE;
	}

	ofp = fopen(of, "w");

	for (i=0; i<vs->cnt; i++) {
		if (vs->s[i].aph.flags & AP_STD_STAR) {
			fstd_indx = i;
			break;
		}
	}
	if (i == vs->cnt) {
		err_printf("Cannot find any standard stars, aborting\n");
		fclose(ofp);
		return ERR_FATAL;
	}

	if (ofp) {
		if (!string_has(vs->repinfo, 'e')) {
			fprintf(ofp, "# cx automated photometry report\n\n");
			report_stars(ofp, vs, NULL, REP_HEADER);
		} else { 
			fprintf(ofp, "# cx automated photometry report\n\n");
			report_event(ofp, vs, NULL, REP_HEADER);
		}
	}

	reset_star_stats();

	iw = new_image_window(0, 0, WIN_IMG, 0);

	if (ac > ind)
		fr = read_fits_file(av[ind]);
	if (fr == NULL)
		return ERR_FILE;
	ret = show_frame(iw, fr, NULL);
	release_frame(fr);

	while(ac > ind && n < COMB_MAX) {
//		d3_printf("RF");
		fr = read_fits_file(av[ind]);
		if (fr != NULL) {
//			info_printf ("file: %s read for stacking\n", av[ind]);
			ret = find_recipy_stars(fr, iw, vs, fstd_indx);
			if (ret && ret != 'a') 
				autop = 0;
			if (ret == 'a')
				autop = 1;
			if (!autop) {
				if (ret == 'n') {
					ind ++;
					continue;
				}
				info_printf("\nPress 'y' to accept frame/star, 'n' to reject frame,\n");
				info_printf("'a' to continue automatically\n");
				do {
					ret = x_select_wait();
//					if (ret == 'p') {
//						dx = iwt[iw].par.star.x - vs->s[fstd_indx].s.x;
//						dy = iwt[iw].par.star.y - vs->s[fstd_indx].s.y;
//						draw_ap_stars(iw, vs, dx, dy);
//					}
					if (ret == 'y' || ret == 'n' || ret == 'a') {
						break;
					}
				} while (1);
				if (ret == 'n') {
					ind ++;
					continue;
				}
				if (ret == 'a') {
					autop = 1;
				}
			} 
			ret = measure_stars(fr, vs);
			if (!ret) {
				get_ph_solution(vs);
				collect_star_stats(vs);
				if (!string_has(vs->repinfo, 'e')) {
					report_stars(stdout, vs, fr, REP_STARS);
					report_stars(ofp, vs, fr, REP_STARS);
				} else { 
					report_event(stdout, vs, fr, REP_STARS);
					report_event(ofp, vs, fr, REP_STARS);
				}
			}
			release_frame(fr);
		} else {
			err_printf("cannot read file: %s, skipping\n", av[ind]);
		}
		ind ++;
	}
	print_noise_summary(stdout, fr, vs);
	report_multistar_stats(stdout, vs);
	if (ofp) {
		if (string_has(vs->repinfo, 'n'))
		    print_noise_summary(ofp, fr, vs);
		if (string_has(vs->repinfo, 's'))
		    report_multistar_stats(ofp, vs);
		fprintf(ofp, "\n# Parameters used for the run\n");
		fprint_recipy(ofp, vs, 2);
		fclose(ofp);
	}
	fflush(stdout);
	return 0;
}


// fix the bad pixels in the files in the list. bf is the name of the badpixel file.
// method is the logical or of the pixel correction and column correction flags (0 is avg)

int fix_files(int ac, char **av, int ind, char *dk, char *ff, char *bf, double mult)
{
	int method = 0;
	char lb[70];
	struct ccd_frame *fr = NULL;
	struct ccd_frame *dkf = NULL;
	struct ccd_frame *fff = NULL;
	struct bad_pix_map map;
	int ret;
	int do_bad = 0;

	map.pixels = 0;
	map.pix = NULL;

	if (dk[0] != 0) {
		dkf = read_fits_file(dk);
		if (dkf == NULL) {
			err_printf("fix_files: cannot load dark frame: %s\n", dk);
			goto err;
		}
		else 
			info_printf("Using dark frame: %s\n", dk);
	}
	if (ff[0] != 0) {
		fff = read_fits_file(ff);
		if (fff == NULL) {
			err_printf("fix_files: cannot load flat field frame: %s\n", ff);
			goto err;
		}
		else 
			info_printf("Using flat field frame: %s\n", ff);
	}
	if (bf[0] != 0) {
		ret = load_bad_pix(&map, bf);
		if (ret < 0) {
// TODO some freeing here
			err_printf("fix_files: cannot read badpix file: %s\n", bf);
			goto err;
		} else 
			info_printf("Using bad pixels map: %s\n", bf);
		do_bad = 1;
	}

	while (ac > ind) {
// load files to process
		fr = read_fits_file(av[ind]);
		if (fr == NULL) {
			err_printf("fix_files: errors reading fits file: %s, skipping\n", av[ind]);
			ind++;
			continue;
		}
// do dark substraction
		if (dkf != NULL) {
			sub_frames (fr, dkf);
			info_printf("dark, ");
			sprintf(lb, "'SUBSTRACTED DARK FRAME %s'", dk);
			fits_add_history(fr, lb);
		}
// do flat fielding
		if (fff != NULL) {
			ret = flat_frame (fr, fff);
			if (ret >= 0) {
				info_printf("flat, ");
				sprintf(lb, "'DIVIDED BY FLAT FRAME %s'", ff);
				fits_add_history(fr, lb);
			}
		}
// bad pixels
		if (do_bad) {
			ret = fix_bad_pixels(fr, &map, method);
			if (ret >= 0) {
				info_printf("bad, ");
				sprintf(lb, "'FIXED %d BAD PIXELS FROM %s'", map.pixels, bf);
				fits_add_history(fr, lb);
			}
		}
// multiply
		if (mult != 0.0) {
			scale_shift_frame(fr, mult, 0.0);
			info_printf("mult, ");
			sprintf(lb, "'MULTIPLIED BY %.6f'", mult);
			fits_add_history(fr, lb);
		}

		info_printf("saving, ");
		write_fits_frame(fr, av[ind]);
		info_printf("done\n");
		release_frame(fr);
		ind ++;
	}
	free(fff);
	free(dkf);
	free(map.pix);
	return 0;
err:
	free(fff);
	free(dkf);
	free(map.pix);
	return ERR_FILE;
}





