/*******************************************************************************
  Copyright(c) 2000 - 2003 Radu Corlan. All rights reserved.
  
  This program is free software; you can redistribute it and/or modify it 
  under the terms of the GNU General Public License as published by the Free 
  Software Foundation; either version 2 of the License, or (at your option) 
  any later version.
  
  This program is distributed in the hope that it will be useful, but WITHOUT 
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
  more details.
  
  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc., 59 
  Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  
  The full GNU General Public License is included in this distribution in the
  file called LICENSE.
  
  Contact Information: radu@corlan.net
*******************************************************************************/

// combine.c: functions used in combining ccd frames

// $Revision: 1.1 $
// $Date: 2003/12/01 00:19:37 $

#include <math.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <time.h>

#include "gcx.h"
#include "x11ops.h"
#include "gsc/gsc.h"

#define APERTURE_RADIUS 15


static int rejected;

// for MMEDIAN, we take the median at each pixel, then average the pixels that are within 
// +/- p1 csigmas * sqrt(median) from it
// increments the pointer array
static double do_mmedian(float *dp[], int n, double csigma, double p1)
{
	int k=0;
	int i;
	double m, d, l1, l2;
	double pix[COMB_MAX];
// copy pixels to the pix array
	for (i=0; i<n; i++) {
		pix[i] = *dp[i];
		dp[i]++;
	}
	m = dmedian(pix, n);
	d = 0.0;
	if (m <= 0.0) {
		l1 = m - 0.5*csigma*p1;
		l2 = m + 0.5*csigma*p1;
	} else {
		l1 = m - sqrt(m)*csigma*p1;
		l2 = m + sqrt(m)*csigma*p1;
	}
// now, average the 'good' pixels
	for (i=0; i<n; i++) {
		if (pix[i] > l1 && pix[i] < l2) {
			k++;
			d += pix[i];
		} else {
			rejected ++;
		}
	}
	return d / k;
}


// for MEDIAN, we simply take the median at each pixel
// increments the pointer array
static double do_median(float *dp[], int n)
{
	int i;
	double pix[COMB_MAX];
// copy pixels to the pix array
	for (i=0; i<n; i++) {
		pix[i] = *dp[i];
		dp[i]++;
	}
	return dmedian(pix, n);
}


// plain averaging
// increments the pointer array
static double do_average(float *dp[], int n)
{
	int i;
	double m=0.0;
// copy pixels to the pix array
	for (i=0; i<n; i++) {
		m += *dp[i];
		dp[i]++;
	}
	return m / n;
}


// stack the n frames in the fr array, with the result in out;
// the frames are aligned at their top-left corner; the frames are stacked
// on the area of the smallest input frame; the rest of the output frame 
// is left unchanged
// returns: the number of frames actually stacked
// p1 and p2 are method-specific parameters
int stack_frames(struct ccd_frame *fr[], struct ccd_frame *out, int n, int method, 
		 double p1, double p2)
{
	double expmin, expmax, totexp;
	int wmin, hmin;
	double csigma, tbias, rnsq, tscale;

	float *dp[COMB_MAX];
	float *op;

	double d;
	int x,y,i;

	char lb[80];

	rejected = 0;

	if (n > COMB_MAX)
		n = COMB_MAX;
// first we collect statistics about the frames
	expmin = HUGE; 
	expmax = 0.0;
	wmin = 32768;
	hmin = 32768;
	totexp=0.0;
	csigma = 0.0;

	d3_printf("collecting frame stats\n");

	for (i = 0; i < n; i++) {
		double exp_time = 1.0;
		if (!fr[i]->stats.statsok)
			frame_stats(fr[i]);
		fits_get_double(fr[i], "EXPTIME", &exp_time);
		if (exp_time > expmax)
			expmax = exp_time;
		if (exp_time < expmin && exp_time > 0) {
			expmin = exp_time;
			csigma = fr[i]->stats.csigma; // this is the worst sigma after scaling
		}
		totexp += exp_time;
		if (fr[i]->w < wmin)
			wmin = fr[i]->w;
		if (fr[i]->h < hmin)
			hmin = fr[i]->h;
	}

	d3_printf("scaling frames\n");

// we now scale the frames with the smaller exposures to 
// match the other (for mmedian and median only)
// we leave frames with unknown exposures unchanged
	if (method == STACK_MMEDIAN || method == STACK_MEDIAN) {
		for (i = 0; i < n; i++) {
			double exp_time = 1.0;
			fits_get_double(fr[i], "EXPTIME", &exp_time);
			if (exp_time != 0 && exp_time < expmax) {
				tscale = expmax / exp_time;
				scale_shift_frame(fr[i], tscale, (1-tscale) * fr[i]->exp.bias);
			}
		}
		csigma *= expmax / expmin;
	}

// check the frame biases
	tscale = 0.0;
	tbias = 0.0;
	rnsq = 0.0;
	for (i = 0; i < n; i++) {
		tbias += fr[i]->exp.bias;
		rnsq += sqr(fr[i]->exp.rdnoise);
		tscale += fr[i]->exp.scale;
	}
	for (i = 0; i < n; i++) {
		if (abs(fr[i]->exp.bias - fr[0]->exp.bias) > 2.0) 
			break;
	}
// compute the noise params
// if all frames are scaled to equal values, the resulting e-/adu
// is the average of the scaled e-/adu's
	switch (method) {
	case STACK_MMEDIAN:
		if (i != n)
			err_printf("Warning: mmedianing frames with different biases!\n");
		out->exp.bias = tbias / n;
		out->exp.rdnoise = sqrt(rnsq) / n;
		out->exp.scale = tscale / n;
		break;
	case STACK_MEDIAN:
		if (i != n)
			err_printf("Warning: medianing frames with different biases!\n");
		out->exp.bias = tbias / n;
		out->exp.rdnoise = 1.2 * sqrt(rnsq) / n;
		out->exp.scale = tscale / n;
		break;
	case STACK_AVERAGE:
		out->exp.bias = tbias / n;
		out->exp.rdnoise = sqrt(rnsq) / n;
		out->exp.scale = tscale / n;
		break;
	default:
		err_printf("stack_frames: unknown stack method: %d\n", method);
		return -1;
	}
	
 

	d3_printf("beginning stack\n");

	for (y = 0; y < hmin; y++) { // lines
		for(i = 0; i < n; i++)
			dp[i] = (float *)(fr[i]->dat) + y * fr[i]->w;
		op = (float *)(out->dat) + y * out->w;
		switch (method) {
		case STACK_MMEDIAN:
			for(x = 0; x < wmin; x++) { // columns
				d = do_mmedian(dp, n, csigma, p1);
				*op = d;
				op ++;
			}
			break;
		case STACK_MEDIAN:
			for(x = 0; x < wmin; x++) { // columns
				d = do_median(dp, n);
				*op = d;
				op ++;
			}
			break;
		case STACK_AVERAGE:
			for(x = 0; x < wmin; x++) { // columns
				d = do_average(dp, n);
				*op = d;
				op ++;
			}
			break;
		default:
			err_printf("stack_frames: unknown stack method: %d\n", method);
			return -1;
		}
	}
// now update the header of the output frame
	sprintf(lb, "%20.2f / TOTAL EXPOSURE TIME", totexp);
	fits_add_keyword(out, "EXPTIME", lb);

	switch (method) {
	case STACK_MMEDIAN:
		sprintf(lb, "'MMEDIAN STACK (REJECT LEVEL %.1f CSIGMAS) OF %d FRAMES'", p1, n);
		fits_add_history(out, lb);
		break;
	case STACK_MEDIAN:
		sprintf(lb, "'MEDIAN STACK OF %d FRAMES'", n);
		fits_add_history(out, lb);
		break;
	case STACK_AVERAGE:
		sprintf(lb, "'AVERAGE STACK OF %d FRAMES'", n);
		fits_add_history(out, lb);
		break;
	}
	return n;

}


// load and combine file with names starting at ind; a new frame is created for the result
// returns NULL in case of error
struct ccd_frame *stack_files(int ac, char **av, int ind, int method, double p)
{
	int n=0;
	struct ccd_frame *ft[COMB_MAX];
	struct ccd_frame *fr, *out=NULL;
	int ret;

	if (ac > ind)
		out = read_fits_file(av[ind]);
	if (out == NULL)
		return NULL;

	while(ac > ind && n < COMB_MAX) {
//		d3_printf("RF");
		fr = read_fits_file(av[ind]);
		if (fr != NULL) {
//			info_printf ("file: %s read for stacking\n", av[ind]);
			ft[n] = fr;
			n++;
		} else {
			err_printf("cannot read file: %s, skipping\n", av[ind]);
		}
		ind ++;
	}

//	d3_printf("RFE ");

	if (n > 0) {
		ret = stack_frames(ft, out, n, method, p, 0.0);
		if (ret == 0)
			return NULL;
		info_printf("successfully combined %d frames. %d pixels rejected\n", 
			    ret, rejected);
		out->stats.statsok = 0;
		return out;
	}
	return NULL;
}



// load and combine file with names starting at ind; a new frame is created for the result
// returns NULL in case of error
// each frame is loaded for display; on the first frame, the program waits for a reference star 
// to be selected; on each successive frame, the user should press y if the suggested ref star 
// is accepted, n if the frame is to be discarded from the stack; if necessary, the program's 
// of the reference star can be changed.

struct ccd_frame *align_stack_files(int ac, char **av, int ind, int method, double p, double f)
{
	int n=0;
	struct ccd_frame *ft[COMB_MAX];
	struct ccd_frame *fr = NULL, *out = NULL;
	int ret;
	struct fr_display_par upar;
	double dx, dy;
	int autop = 0;

	float kern[100];
	int kern_size;

	double sxmin, sxmax, symin, symax;
	int w,h,sx,sy;
	int iw;

	sxmin=symin=HUGE;
	sxmax=symax=-HUGE;

	kern_size = 2 * floor(f) + 1;
	if (kern_size < 3)
		kern_size = 3;
	if (kern_size > 7)
		kern_size = 7;

	ret = make_gaussian(0.4 * f, kern_size, kern);
	if (ret < 0) {
		err_printf("align_stack_files: cannot make gaussian kernel, aborting\n");
		return NULL;
	}

	iw = new_image_window(0, 0, WIN_IMG, 0);

	if (ac > ind)
		out = read_fits_file(av[ind]);
	if (out == NULL)
		return NULL;

	ret = show_frame(iw, out, NULL);
	release_frame(fr);
	do {
		info_printf("\nPress 'o' over refernce star to select it\n");
		while ((ret = x_select_wait()) != 'o') ;
	} while (iwt[iw].par.ref_star.x == 0 && iwt[iw].par.ref_star.y == 0);

	get_star_near(out, (int)iwt[iw].par.ref_star.x, (int)iwt[iw].par.ref_star.y, 
		      0, &(iwt[iw].par.star));

	while(ac > ind && n < COMB_MAX) {
//		d3_printf("RF");
		fr = read_fits_file(av[ind]);
		if (fr != NULL) {
//			info_printf ("file: %s read for stacking\n", av[ind]);
			memcpy(&upar, &iwt[iw].par, sizeof(struct fr_display_par) );
			upar.map_changed = 1;
			follow_star(fr, 20.0, &upar.star, &upar.star);
			ret = show_frame(iw, fr, &upar);

			if (!autop) {
				info_printf("\nPress 'y' to accept frame/star, 'n' to reject frame,\n");
				info_printf("'p' to change star selection or 'a' to continue unattended\n");
				while (!(ret = x_select_wait())) ;
				if (ret == 'n') {
					ind ++;
					continue;
				}
				if (ret == 'a')
					autop = 1;
			} 
			if (autop) { // check FWHM
				if (iwt[iw].par.star.fwhm > f) {
					info_printf("FWHM too large, rejecting frame\n");
					ind ++;
					continue;
				}
			}

			dx = iwt[iw].par.ref_star.x - iwt[iw].par.star.x;
			dy = iwt[iw].par.ref_star.y - iwt[iw].par.star.y;
			info_printf("frame shift: %.2f %.2f filter sigma: %.1f\n", 
				    dx, dy, 0.3 * f);

			filter_frame_inplace(fr, kern, kern_size);

			shift_frame(fr, dx, dy);
			if (dx < sxmin)
				sxmin = dx;
			if (dy < symin)
				symin = dy;
			if (dx > sxmax)
				sxmax = dx;
			if (dy > symax)
				symax = dy;
			ft[n] = fr;
			n++;
		} else {
			err_printf("cannot read file: %s, skipping\n", av[ind]);
		}
		ind ++;
	}

//	d3_printf("RFE ");

	if (n > 0) {
		ret = stack_frames(ft, out, n, method, p, 0.0);
		if (ret == 0)
			return NULL;
		info_printf("successfully combined %d frames. %d pixels rejected\n", 
			    ret, rejected);
		out->stats.statsok = 0;
		w = out->w - (int)ceil(2*(sxmax - sxmin));
		h = out->h - (int)ceil(2*(symax - symin));
		sx = (int)ceil(sxmax);
		sy = (int)ceil(symax);
		info_printf("new frame size: %d %d from %d %d\n", w, h, sx, sy);
		crop_frame(out, sx, sy, w, h);
		return out;
	}
	return NULL;
}


// align files in list to the first one
// each frame is loaded for display; on the first frame, the program waits for a reference star 
// to be selected; on each successive frame, the user should press y if the suggested ref star 
// is accepted, n if the frame is to be skipped; if necessary, the program's choice 
// of the reference star can be changed.

int align_files(int ac, char **av, int ind)
{
	int n=0;
	char lb[70];
	struct ccd_frame *fr, *out = NULL;
	int ret;
	struct fr_display_par upar;
	double dx, dy;
	int autop = 0;

	int iw;

	iw = new_image_window(0, 0, WIN_IMG, 0);

	if (ac > ind)
		out = read_fits_file(av[ind]);
	if (out == NULL)
		return ERR_FILE;

	ret = show_frame(iw, out, NULL);
	do {
		info_printf("\nPress 'o' over refernce star to select it\n");
		while ((ret = x_select_wait()) != 'o') ;
	} while (iwt[iw].par.ref_star.x == 0 && iwt[iw].par.ref_star.y == 0);

	get_star_near(out, (int)iwt[iw].par.ref_star.x, (int)iwt[iw].par.ref_star.y, 
		      0, &(iwt[iw].par.star));

	ind ++;

	while(ac > ind && n < COMB_MAX) {
//		d3_printf("RF");
		fr = read_fits_file(av[ind]);
		if (fr != NULL) {
//			info_printf ("file: %s read for stacking\n", av[ind]);
			memcpy(&upar, &iwt[iw].par, sizeof(struct fr_display_par) );
			upar.map_changed = 1;
			follow_star(fr, 20.0, &upar.star, &upar.star);

			ret = show_frame(iw, fr, &upar);

			if (!autop) {
				info_printf("\nPress 'y' to accept frame/star, 'n' to reject frame,\n");
				info_printf("'p' to change star selection or 'a' to continue unattended\n");
				while (!(ret = x_select_wait())) ;
				if (ret == 'n') {
					ind ++;
					continue;
				}
				if (ret == 'a')
					autop = 1;
			} 

			dx = iwt[iw].par.ref_star.x - iwt[iw].par.star.x;
			dy = iwt[iw].par.ref_star.y - iwt[iw].par.star.y;
			info_printf("frame shift: %.2f %.2f\n", dx, dy);
			shift_frame(fr, dx, dy);
			sprintf(lb, "'FRAME ALIGNED DX=%.2f DX=%.2f'", dx, dy);
			fits_add_history(fr, lb);
			write_fits_frame(fr, av[ind]);
			release_frame(fr);
		} else {
			err_printf("cannot read file: %s, skipping\n", av[ind]);
		}
		ind ++;
	}
	return 0;

}




// return a method code from the supplied method string
// returns MMEDIAN if an invalid method is supplied
int parse_stack_method(char *opt)
{
	if (!strncmp(opt, "mm", 2))
		return STACK_MMEDIAN;
	if (!strncmp(opt, "me", 2))
		return STACK_MEDIAN;	
	if (!strncmp(opt, "av", 2))
		return STACK_AVERAGE;
	err_printf("Unknown method %s, assuming average\n", opt);
	return STACK_AVERAGE;
}

// remove duplicates from match
// this is n**2 code, hopefully won't be called for large datasets
int starmatch_duplicates (int *cix, int nm)
{
	int i, j;
	for (i = 0; i < nm; i++)
		for (j = i + 1; j < nm; j++) {
			if (cix[i] == cix[j]) {
				cix[i] = cix[j] = -1;
			}
		}
// count remaining matches
	j = 0;
	for (i = 0; i < nm; i++)
		if (cix[i] >= 0)
			j++;
	return j;

}

int starmatch_wrapper(float *sx, float *sy, float *sm, int sn,    // input sources in picture
              float *cx, float *cy, float *cm, int cn,    // input catalog
              int *cix, float *omega, float *dx, float *dy, 
              float *ampx, float *ampy,
              float *dmag)  // out; cix is a vector [sn]
{
	int nm = 0, ret, i;
	FILE *sfp;

	sfp = fopen("starmatch_deb.out", "a");
	if (sfp == NULL)
		exit(0);

	fprintf(sfp, "Starmatch run\n");
	fprintf(sfp, "sx sy sm (sn=%d)\n", sn);
	for(i = 0; i < sn; i++)
		fprintf(sfp, "%8.2f %8.2f %8.2f\n", sx[i], sy[i], sm[i]);
	fprintf(sfp, "cx cy cm (cn=%d)\n", cn);
	for(i = 0; i < cn; i++)
		fprintf(sfp, "%8.2f %8.2f %8.2f\n", cx[i], cy[i], cm[i]);

	fprintf(sfp, "dx: %.1f, dy:%.1f, omega:%.3f, ampx:%.1f, ampy: %.1f\n",
		  *dx, *dy, *omega, *ampx, *ampy);

//	nm = starmatch(sx, sy, sm, sn, cx, cy, cm, cn, cix, 
//			omega, dx, dy, ampx, ampy, dmag);

	ret = starmatch_duplicates(cix, cn);

	fprintf(sfp, "starmatch_wrapper: matched %d stars, after deduping %d\n", nm, ret);

	for (i=0; i<cn; i++) {
		if (cix[i] > 0)
			fprintf(sfp, "cix[%d] (%.1f,%.1f) = %d (%.1f,%.1f)\n", 
				i, sx[i], sy[i], cix[i], cx[cix[i]], cy[cix[i]]);
	}

	fclose(sfp);

	return ret;
}

#define MATCH_SOURCES 20 // max nr of sources to use for matching

// find the transformation matrix to go from ssrc coord to csrc coords so as to match stars
// the order of the transform in ct is never increased
// returns the number of stars successfully used for the transform, or a negative error code

static int find_transform(struct sources *ssrc, struct sources *csrc, struct ctrans *ct)
{
	float *sx, *sy, *sm, *cx, *cy, *cm;
	int sn, cn;
	int *cix;
	float omega, dx, dy, ampx, ampy, dmag;
	int i, ret;

// check we have enough stars

	sn = ssrc->ns < MATCH_SOURCES ? ssrc->ns : MATCH_SOURCES;
	cn = csrc->ns < MATCH_SOURCES ? csrc->ns : MATCH_SOURCES;

	if (sn < 3 || cn < 3) {
		err_printf("find_transform: need at least three stars to attempt match\n");
		return -1;
	}

	d3_printf("sn=%d cn=%d\n", sn, cn);

// alloc data for starmatch

	sx = sy = sm = cx = cy = cm = (float *)NULL;
	cix = (int *)NULL;

	sx = calloc(sn, sizeof(float));
	if (!sx) goto err;
	sy = calloc(sn, sizeof(float));
	if (!sy) goto err;
	sm = calloc(sn, sizeof(float));
	if (!sm) goto err;
	cx = calloc(cn, sizeof(float));
	if (!cx) goto err;
	cy = calloc(cn, sizeof(float));
	if (!cy) goto err;
	cm = calloc(cn, sizeof(float));
	if (!cm) goto err;
	cix = calloc(cn, sizeof(int));
	if (!cix) goto err;

// fill them up with data
	for (i = 0; i < sn; i++) {
		sx[i] = ssrc->s[i].x;
		sy[i] = ssrc->s[i].y;
		sm[i] = -2.5 * log10(ssrc->s[i].flux);
	}
	for (i = 0; i < cn; i++) {
		cx[i] = csrc->s[i].x;
		cy[i] = csrc->s[i].y;
		cm[i] = -2.5 * log10(csrc->s[i].flux);
	}

// set parameters
	omega = 0;
	dx = 0;
	dy = 0;
	ampx = 1;
	ampy = 1;
	dmag = 0;

	d3_printf("calling starmatch with sn=%d cn=%d\n", sn, cn);

	ret = starmatch_wrapper(sx, sy, sm, sn, cx, cy, cm, cn, cix, 
			&omega, &dx, &dy, &ampx, &ampy, &dmag);

	d3_printf("find_transform: matched %d stars\n", ret);
	if (ret <= 0)
		goto err;

	d3_printf("dx: %.1f, dy:%.1f, omega:%.3f, ampx:%.1f, ampy: %.1f\n",
		  dx, dy, omega, ampx, ampy);

	for (i=0; i<ret && i <10; i++) {
		if (cix[i] >= 0) 
			d3_printf("cix[%d] (%.1f,%.1f) = %d (%.1f,%.1f)\n", 
				  i, sx[i], sy[i], cix[i], cx[cix[i]], cy[cix[i]]);
	}

// free data and return
err:
	free(sx);
	free(sy);
	free(sm);
	free(cx);
	free(cy);
	free(cm);
	free(cix);
	return -2;
}  


// align files in list to the first one
// using full-field starmatching

int full_align_files(int ac, char **av, int ind)
{
	int n=0;
	char lb[70];
	struct ccd_frame *fr, *out=NULL;
	int ret;
	double dx, dy;
	int autop = 0;

	struct ctrans ct;

	int iw;

	struct star *cst;
	struct sources csrc;

	struct star *sst;
	struct sources ssrc;

	iw = new_image_window(0, 0, WIN_IMG, 0);

	if (ac > ind)
		out = read_fits_file(av[ind]);
	if (out == NULL)
		return ERR_FILE;

//	ret = show_frame(iw, out, NULL);

// allocate array for sources
	cst = calloc(MATCH_SOURCES, sizeof(struct star));
	sst = calloc(MATCH_SOURCES, sizeof(struct star));
	if (cst == NULL || sst == NULL) {
		err_printf("cannot alloc\n");
		free(cst);
		free(sst);
		return -1;
	}
	csrc.ns = 0;
	csrc.maxn = MATCH_SOURCES;
	csrc.s = cst;
	ssrc.ns = 0;
	ssrc.maxn = MATCH_SOURCES;
	ssrc.s = sst;

	ret = extract_stars(out, NULL, 0, 9, &csrc);

	ind ++;

	while(ac > ind && n < COMB_MAX) {
//		d3_printf("RF");
		fr = read_fits_file(av[ind]);
		if (fr != NULL) {
//			info_printf ("file: %s read for stacking\n", av[ind]);
//			memcpy(&upar, &iwt[iw].par, sizeof(struct fr_display_par) );
//			upar.map_changed = 1;
//			follow_star(fr, 20.0, &upar.star, &upar.star);

			ret = show_frame(iw, fr, NULL);
			ret = extract_stars(fr, NULL, 0, 9, &ssrc);

			show_sources(iw, &ssrc);

			ret = find_transform(&ssrc, &csrc, &ct);

			ssrc.ns = 0;

			if (!autop) {
				info_printf("\nPress 'y' to accept frame/star, 'n' to reject frame,\n");
				info_printf("'p' to change star selection or 'a' to continue unattended\n");
				while (!(ret = x_select_wait())) ;
				if (ret == 'n') {
					ind ++;
					continue;
				}
				if (ret == 'a')
					autop = 1;
			} 

			dx = iwt[iw].par.ref_star.x - iwt[iw].par.star.x;
			dy = iwt[iw].par.ref_star.y - iwt[iw].par.star.y;
			info_printf("frame shift: %.2f %.2f\n", dx, dy);
//			shift_frame(fr, dx, dy);
			sprintf(lb, "'FRAME ALIGNED DX=%.2f DX=%.2f'", dx, dy);
//			fits_add_history(fr, lb);
//			write_fits_frame(fr, av[ind]);
			release_frame(fr);
		} else {
			err_printf("cannot read file: %s, skipping\n", av[ind]);
		}
		ind ++;
	}
	return 0;

}
