/*******************************************************************************
  Copyright(c) 2000 - 2003 Radu Corlan. All rights reserved.
  
  This program is free software; you can redistribute it and/or modify it 
  under the terms of the GNU General Public License as published by the Free 
  Software Foundation; either version 2 of the License, or (at your option) 
  any later version.
  
  This program is distributed in the hope that it will be useful, but WITHOUT 
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
  more details.
  
  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc., 59 
  Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  
  The full GNU General Public License is included in this distribution in the
  file called LICENSE.
  
  Contact Information: radu@corlan.net
*******************************************************************************/

// gcx.c: main file for camera control program 
// $Revision: 1.11 $
// $Date: 2004/01/11 02:48:18 $

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/io.h>
#include <sys/time.h>
#include <unistd.h>
#include <getopt.h>
#include <pwd.h>
#include <errno.h>

#include "gcx.h"
#include "x11ops.h"
#include "catalogs.h"
#include "gui.h"
#include "params.h"
#include "filegui.h"
#include "obslist.h"
#include "helpmsg.h"
#include "recipy.h"


void show_usage(void) {
	info_printf("%s", help_usage_page);
//	info_printf("\nFrame stacking options\n");
//	info_printf("-o <ofile>   Set output file name (for -c, -s, -a, -p) \n");
//	info_printf("-c <method>  combine files in the file list\n");
//	info_printf("             using the specified method. Valid methods\n");
//	info_printf("             are: mm(edian), av(erage), me(dian)\n");
//	info_printf("-s <method>  Stack (align and combine) files in the file\n");
//	info_printf("             list using same methods as -c\n");
//	info_printf("-a           align files in-place\n");
//	info_printf("-w <value>   Max FWHM accepted when auto-stacking frames [5.0]\n");
//	info_printf("\nFrame processing options\n");
//	info_printf("-d <dkfile>  Substract <dkfile> from files\n");
//	info_printf("-f <fffile>  Flatfield files using <fffile> (maintain avg flux)\n");
//	info_printf("-b <bfile>   Fix bad pixels in fits files using <bfile>\n");
//	info_printf("-m <mult>    Multiply files by the scalar <mult>\n");
//	info_printf("Any of -d, -f, -b and -m can be invoked at the same time; in this\n");
//	info_printf("    case, the order of processing is dark, flat, badpix, multiply\n");
//	info_printf("    These options will cause the frame stack options to be ingnored\n");
}

static int load_par_file(char *fn, GcxPar p)
{
	FILE *fp;
	int ret;

	fp = fopen(fn, "r");
	if (fp == NULL) {
		err_printf("load_par_file: cannot open %s\n", fn);
		return -1;
	}

	ret = fscan_params(fp, p);
	fclose(fp);
	return ret;
}

static int save_par_file(char *fn, GcxPar p)
{
	FILE *fp;
	fp = fopen(fn, "w");
	if (fp == NULL) {
		err_printf("save_par_file: cannot open %s\n", fn);
		return -1;
	}
	fprint_params(fp, p);
	fclose(fp);
	return 0;
}

static int load_params_rc(void)
{
	uid_t my_uid;
	struct passwd *passwd;
	char *rcname;
	int ret;

	my_uid = getuid();
	passwd = getpwuid(my_uid);
	if (passwd == NULL) {
		err_printf("load_params_rc: cannot determine home directoy\n");
		return -1;
	}
	asprintf(&rcname, "%s/.gcxrc", passwd->pw_dir);
	ret = load_par_file(rcname, PAR_NULL);
	free(rcname);
	return ret;
}

int save_params_rc(void)
{
	uid_t my_uid;
	struct passwd *passwd;
	char *rcname;
	int ret;

	my_uid = getuid();
	passwd = getpwuid(my_uid);
	if (passwd == NULL) {
		err_printf("save_params_rc: cannot determine home directoy\n");
		return -1;
	}
	asprintf(&rcname, "%s/.gcxrc", passwd->pw_dir);
	ret = save_par_file(rcname, PAR_NULL);
	free(rcname);
	return ret;
}

static void recipy_file_convert(char *rcf, char *outf)
{
	FILE *infp = NULL, *outfp = NULL;
	int nc;
	if (rcf[0] != 0) {
		infp = fopen(rcf, "r");
		if (infp == NULL) {
			err_printf("Cannot open file %s for reading\n%s\n", 
				   rcf, strerror(errno));
			exit(1);
		}
	}
	if (outf[0] != 0) {
		outfp = fopen(outf, "w");
		if (infp == NULL) {
			err_printf("Cannot open file %s for writing\n%s\n", 
				   outf, strerror(errno));
			fclose(infp);
			exit(1);
		}
	}
	if (outfp && infp)
		nc = convert_recipy(infp, outfp);
	else if (outfp)
		nc = convert_recipy(stdin, outfp);
	else if (infp)
		nc = convert_recipy(infp, stdout);
	else
		nc = convert_recipy(stdin, stdout);

	if (nc < 0) {
		err_printf("Error converting recipy\n");
		if (infp)
			fclose(infp);
		if (outfp)
			fclose(outfp);
		exit(1);
	} else {
		if (infp)
			fclose(infp);
		if (outfp) {
			info_printf("%d star(s) written\n", nc);
			fclose(outfp);
		}
		exit (0);
	}
}

static void report_convert(char *rcf, char *outf)
{
	FILE *infp = NULL, *outfp = NULL;

	if (rcf[0] != 0) {
		infp = fopen(rcf, "r");
		if (infp == NULL) {
			err_printf("Cannot open file %s for reading\n%s\n", 
				   rcf, strerror(errno));
			exit(1);
		}
	}
	if (outf[0] != 0) {
		outfp = fopen(outf, "w");
		if (infp == NULL) {
			err_printf("Cannot open file %s for writing\n%s\n", 
				   outf, strerror(errno));
			if (infp)
				fclose(infp);
			exit(1);
		}
	}
	if (outfp && infp)
		report_to_table(infp, outfp, NULL);
	else if (infp)
		report_to_table(infp, stdout, NULL);
	else if (outfp)
		report_to_table(stdin, outfp, NULL);
	else
		report_to_table(stdin, stdout, NULL);
	if (infp)
		fclose(infp);
	if (outfp) 
		fclose(outfp);
	exit (0);
}


static void landolt_file_convert(char *rcf, char *outf)
{
	FILE *infp, *outfp = NULL;
	int nc;
	infp = fopen(rcf, "r");
	if (infp == NULL) {
		err_printf("Cannot open file %s for reading\n%s\n", 
			   rcf, strerror(errno));
		exit(1);
	}
	if (outf[0] != 0) {
		outfp = fopen(outf, "w");
		if (infp == NULL) {
			err_printf("Cannot open file %s for writing\n%s\n", 
				   outf, strerror(errno));
			fclose(infp);
			exit(1);
		}
	}
	if (outfp)
		nc = convert_landolt(infp, outfp);
	else 
		nc = convert_landolt(infp, stdout);
	if (nc < 0) {
		err_printf("Error converting landolt table\n");
		fclose(infp);
		if (outfp)
			fclose(outfp);
		exit(1);
	} else {
		fclose(infp);
		if (outfp) {
			info_printf("%d star(s) written\n", nc);
			fclose(outfp);
		}
		exit (0);
	}
}




int main(int ac, char **av)
{
	int ret;
	char *file;
	struct ccd_frame *fr, *out;
	char fn[1024]="out";
	char bf[1024]="";
	char dk[1024]="";
	char ff[1024]="";
	char rf[1024]="";
	char repf[1024]=""; /* report file to be converted */
	char ldf[1024]=""; /* landolt table file to be converted */
	char rcf[1024]=""; /* recipy file to be converted */
	char outf[1024]=""; /* outfile */
	char of[1024]=""; /* obsfile */
	int to_pnm = 0;
	int run_phot = 0;
	int convert_recipy = 0;
	int rep_to_table = 0;
	char *rc_fn = NULL;
	int iw;
	char oc;
	float p=1.0;
	float f=5.0;
	int interactive = 0;
	int batch = 0;
	int method = -1;
	int amethod = -1;
	int align = 0;
	int full_align = 0;
//	struct fr_display_par upar;
	int firstind;
	int badfix = 0;
	int flatproc = 0;
	int darksub = 0;
	float mult = 0.0;
	GtkWidget *window;

	char *shortopts = "r:p:hP:vo:id:b:f:B:M:O:usaT:S:n";
	struct option longopts[] = {
		{"dark", required_argument, NULL, 'd'},
		{"bias", required_argument, NULL, 'b'},
		{"flat", required_argument, NULL, 'f'},
		{"badpix", required_argument, NULL, 'B'},
		{"add-bias", required_argument, NULL, 'A'},
		{"multiply", required_argument, NULL, 'M'},

		{"output", required_argument, NULL, 'o'},
		{"update-file", no_argument, NULL, 'u'},

		{"stack", no_argument, NULL, 's'},
		{"align", no_argument, NULL, 'a'},

		{"set", required_argument, NULL, 'S'},

		{"rcfile", required_argument, NULL, 'r'},
		{"obsfile", required_argument, NULL, 'O'},
		{"recipy", required_argument, NULL, 'p'},
		{"convert-rcp", required_argument, NULL, '2'},
		{"to-pnm", no_argument, NULL, 'n'},
		{"convert-landolt", required_argument, NULL, 'L'},
		{"rep-to-table", required_argument, NULL, 'T'},
		{"phot-run", required_argument, NULL, 'P'},
		{"interactive", no_argument, NULL, 'i'},
		{"help", no_argument, NULL, 'h'},
		{"version", no_argument, NULL, '1'},
		{NULL, 0, NULL, 0}
	};

	init_ptable();
	load_params_rc();

	while ((oc = getopt_long(ac, av, shortopts, longopts, NULL)) > 0) {
		switch(oc) {
		case 'h':
			show_usage();
			exit(0);
		case '1':
			info_printf("%s\n", VERSION);
			exit(0);
		case ':':
			err_printf("option -%c requires an argument\n", optopt);
			exit(1);
		case '?':
			err_printf("unknown option -%c\n", optopt);
			exit(1);
		case 'i':
			interactive = 1;
			break;
		case 'n':
			to_pnm = 1;
			batch = 1;
			break;
		case 'O':
			strncpy(of, optarg, 1023);
			break;
		case 'o':
			strncpy(outf, optarg, 1023);
			break;
		case 'p':
			sscanf(optarg, "%800s", rf);
			break;
		case 'P':
			sscanf(optarg, "%800s", rf);
			run_phot = 1;
			batch = 1;
			break;
		case '2':
			if (optarg) {
				if (optarg[0] == '-' && optarg[1] == 0)
					;
				else
					strncpy(rcf, optarg, 1023);
			}
			convert_recipy = 1;
			batch = 1;
			break;
		case 'S':
			if (optarg) {
				if (set_par_by_name(optarg))
					err_printf("error setting %s\n", optarg);
			}
			break;
		case 'T':
			if (optarg) {
				if (optarg[0] == '-' && optarg[1] == 0)
					;
				else
					strncpy(repf, optarg, 1023);
			}
			rep_to_table = 1;
			batch = 1;
			break;
		case 'L':
			d3_printf("ldf\n");
			if (optarg)
				strncpy(ldf, optarg, 1023);
			break;
		case 'c':
			method = parse_stack_method(optarg);
			break;
		case 's':
			amethod = parse_stack_method(optarg);
			break;
		case 'a':
			align = 1;
			break;
		case 'A':
			full_align = 1;
			break;
		case 'r':
			sscanf(optarg, "%f", &p);
			break;
		case 'w':
			sscanf(optarg, "%f", &f);
			break;
		case 'b':
			sscanf(optarg, "%800s", bf);
			badfix = 1;
			break;
		case 'f':
			sscanf(optarg, "%800s", ff);
			flatproc = 1;
			break;
		case 'd':
			sscanf(optarg, "%800s", dk);
			darksub = 1;
			break;
		case 'm':
			sscanf(optarg, "%f", &mult);
			break;
		}
	}


	if (rc_fn != NULL) {
		load_par_file(rc_fn, PAR_NULL);
	}

	gtk_init (&ac, &av);

	if (convert_recipy) { /* convert recipy */
		recipy_file_convert(rcf, outf);
	}

	if (rep_to_table) { /* convert a report into tabular format */
		report_convert(repf, outf);
	}

	if (ldf[0] != 0) { /* convert landolt table */
		d3_printf("landolt convert\n");
		landolt_file_convert(ldf, outf);
	}

	if (of[0] != 0) {
		if ((strchr(of, '/') == NULL)) {
			file = find_file_in_path(of, P_STR(FILE_OBS_PATH));
			if (file != NULL) {
				strncpy(of, file, 1023);
				of[1023] = 0;
				free(file);
			}
		}
		d3_printf("of is %s\n", of);
	}

	if (rf[0] != 0) {
		if ((strchr(of, '/') == NULL)) {
			file = find_file_in_path(rf, P_STR(FILE_RCP_PATH));
			if (file != NULL) {
				strncpy(rf, file, 1023);
				rf[1023] = 0;
				free(file);
			}
		}
		d3_printf("rf is %s\n", rf);
	}


	if (ac == 1)
		interactive = 1;

	if (badfix || flatproc || darksub || mult != 0.0) {
		fix_files(ac, av, optind, dk, ff, bf, mult);
		exit(0);
	}

//	xops_init();

	if (method >= 0) { // we do some stacking
		out = stack_files(ac, av, optind, method, p);
		if (out != NULL) {
			write_fits_frame(out, fn);
			iw = new_image_window(0, 0, WIN_IMG, 0);
			ret = show_frame(iw, out, NULL);
			info_printf("\nPress 'q' to exit program\n");
			while ( (ret = x_select_wait()) != 'q') ;
			exit (0);
		}
	} else if (amethod >= 0) { // we do some stacking
		out = align_stack_files(ac, av, optind, amethod, p, f);
		if (out != NULL) {
			write_fits_frame(out, fn);
			iw = new_image_window(0, 0, WIN_IMG, 0);
			ret = show_frame(iw, out, NULL);
			info_printf("\nPress 'q' to exit program\n");
			while ( (ret = x_select_wait()) != 'q') ;
			exit (0);
		}
	} else if (align) {
		align_files(ac, av, optind);
		exit(0);
	} /* else if (full_align) {
		full_align_files(ac, av, optind);
		exit(0);
		} */

	firstind = optind;

	window = create_image_window();
	if (!batch || interactive) {
		gtk_widget_show_all(window);
	}

	if (ac > optind) {
		fr = read_gz_fits_file(av[optind], P_STR(FILE_UNCOMPRESS));
		if (fr == NULL) {
			err_printf("cannot read file %s", av[optind]);
			exit(1);
		}

		frame_to_channel(fr, window, "i_channel");
		release_frame(fr);
		if (to_pnm) {
			struct image_channel *channel;
			channel = gtk_object_get_data(GTK_OBJECT(window), "i_channel");
			if (channel == NULL) {
				err_printf("oops - no channel\n");
				if (batch && !interactive)
					exit(1);
			}
			if (outf[0] != 0) {
				ret = channel_to_pnm_file(channel, NULL, outf);
			} else {
				ret = channel_to_pnm_file(channel, NULL, NULL);
			}
			if (ret) {
				err_printf("error writing pnm file\n");
				if (batch && !interactive)
					exit(1);
			}
			if (batch && !interactive)
				exit(0);
		}
		if (rf[0] != 0) {
			FILE *output_file = NULL;
			load_rcp_to_window(window, rf);
			match_field_in_window(window);
			if (run_phot) {
				if (outf[0] != 0) {
					output_file = fopen(outf, "a");
					if (output_file == NULL)
						err_printf("Cannot open file %s for appending\n%s\n", 
							   outf, strerror(errno));
				}
				if (output_file == NULL)
					phot_to_fd(window, stdout);
				else {
					phot_to_fd(window, output_file);
					fclose(output_file);
				}
			}
			if (batch && !interactive) {
				exit(0);
			}
		}
		d3_printf("calling gtk_main\n");
		gtk_main ();
		exit(0);
	}
	d3_printf("calling gtk_main, no frame\n");
	if (of[0] != 0) { /* run obs */
		if (run_obs_file(window, of))
			exit (1);
	}
	gtk_main ();
	exit(0);
}
















