/*******************************************************************************
  Copyright(c) 2000 - 2003 Radu Corlan. All rights reserved.
  
  This program is free software; you can redistribute it and/or modify it 
  under the terms of the GNU General Public License as published by the Free 
  Software Foundation; either version 2 of the License, or (at your option) 
  any later version.
  
  This program is distributed in the hope that it will be useful, but WITHOUT 
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
  more details.
  
  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc., 59 
  Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  
  The full GNU General Public License is included in this distribution in the
  file called LICENSE.
  
  Contact Information: radu@corlan.net
*******************************************************************************/
#define _GNU_SOURCE

#include <stdio.h>
#include <malloc.h>
#include <math.h>
#include <string.h>
#include <ctype.h>

#include "gcx.h"
#include "params.h"
#include "catalogs.h"
#include "sourcesdraw.h"

struct param ptable[PAR_TABLE_SIZE];

static char * star_shapes[] = {
	"circle", "square", "blob", "aphot", "diamond", "cross", NULL
};

static char * color_types[] = {
	"instrumental", "standard", NULL
};

static char * sky_methods[] = {
	"average", "median", "kappa-sigma", "mean-median", "synthetic-mode", NULL
};

/* functions used for parameter table initialisation */

/* find an attachemnt node (last of parent's children)
 * return -1 if no suitable point is found, or the node.
 * a link to the new item is added in the found node
 */
static int add_link(GcxPar parent, GcxPar p)
{
	int point;
	if (parent == PAR_NULL) {
		point = PAR_FIRST;
		while (PAR(point)->next != PAR_NULL)
			point = PAR(point)->next;
		PAR(point)->next = p;
		return point;
	}
	if (PAR_TYPE(parent) != PAR_TREE) {
		err_printf("add_par: parent %d is not a tree!\n", parent);
		return -1;
	}
	if (PAR(parent)->child == PAR_NULL) {
		PAR(parent)->child = p;
		return parent;
	}
	point = PAR(parent)->child;
	while (PAR(point)->next != PAR_NULL)
		point = PAR(point)->next;
//	d3_printf("chaining to %d\n", point);
	PAR(point)->next = p;
	return point;
}

static void add_par_tree(GcxPar p, GcxPar parent, char *name, char *comment)
{
	int point;

	if (p >= PAR_TABLE_SIZE || parent >= PAR_TABLE_SIZE) {
		err_printf("add_par: bad par[%d] index \n", p);
		return;
	}
	point = add_link(parent, p);
	if (point <= 0) {
		err_printf("add_par: cannot find link point for [%d]\n", p);
		return;
	}
	PAR(p)->parent = (parent);
	PAR(p)->child = PAR_NULL;
	PAR(p)->next = PAR_NULL;
	PAR(p)->flags = PAR_TREE;
	PAR(p)->name = name;
	PAR(p)->comment = comment;
}

static void add_par_int(GcxPar p, GcxPar parent, int format, char *name, char *comment, int v)
{
	int point;

	if (p >= PAR_TABLE_SIZE || parent >= PAR_TABLE_SIZE) {
		err_printf("add_par: bad par index\n");
		return;
	}
	point = add_link(parent, p);
	if (point <= 0) {
		err_printf("add_par: cannot find link point for [%d]\n", p);
		return;
	}
	PAR(p)->parent = (parent);
	PAR(p)->child = (PAR_NULL);
	PAR(p)->next = PAR_NULL;
	PAR(p)->flags = PAR_INTEGER | format;
	PAR(p)->name = name;
	PAR(p)->comment = comment;
	P_INT(p) = v;
	PAR(p)->defval.i = v;
}

static void add_par_double(GcxPar p, GcxPar parent, int format, char *name, char *comment, 
			   double v)
{
	int point;

	if (p >= PAR_TABLE_SIZE || parent >= PAR_TABLE_SIZE) {
		err_printf("add_par: bad par index\n");
		return;
	}
	point = add_link(parent, p);
	if (point <= 0) {
		err_printf("add_par: cannot find link point for [%d]\n", p);
		return;
	}
	PAR(p)->parent = (parent);
	PAR(p)->child = (PAR_NULL);
	PAR(p)->next = PAR_NULL;
	PAR(p)->flags = PAR_DOUBLE | format;
	PAR(p)->name = name;
	PAR(p)->comment = comment;
	P_DBL(p) = v;
	PAR(p)->defval.d = v;
}

static void add_par_string(GcxPar p, GcxPar parent, int format, char *name, char *comment, 
			   char *val)
{
	int point;

//	d3_printf("add_par_string %s %d [%d]\n", name, p, parent);

	if (p >= PAR_TABLE_SIZE || parent >= PAR_TABLE_SIZE) {
		err_printf("add_par: bad par index\n");
		return;
	}
	point = add_link(parent, p);
	if (point <= 0) {
		err_printf("add_par: cannot find link point for [%d]\n", p);
		return;
	}
	PAR(p)->parent = (parent);
	PAR(p)->child = PAR_NULL;
	PAR(p)->next = PAR_NULL;
	PAR(p)->flags = PAR_STRING | format;
	PAR(p)->name = name;
	PAR(p)->comment = comment;
	if (P_STR(p) != NULL) {
		free(P_STR(p));
		P_STR(p) = NULL;
	}
	if (PAR(p)->defval.s != NULL) {
		free(PAR(p)->defval.s);
		PAR(p)->defval.s = NULL;
	}
	P_STR(p) = malloc(strlen(val)+1);
	if (P_STR(p) != NULL)
		strcpy(P_STR(p), val);
	PAR(p)->defval.s = malloc(strlen(val)+1);
	if (PAR(p)->defval.s != NULL)
		strcpy(PAR(p)->defval.s, val);
}

/*
 * make an int parameter show as a choices list
 * choices must be valid for the lifetime of the
 * program; static constants are good*/
static void set_par_choices(GcxPar p, char ** choices)
{
	int i = 0;
	if (PAR_TYPE(p) != PAR_INTEGER) {
		err_printf("cannot add choices to a non-int par\n");
		return;
	}
	PAR(p)->choices = choices;
	PAR(p)->flags &= ~PAR_FORMAT_FM;
	PAR(p)->flags |= FMT_OPTION;
	while (choices[i] != NULL) 
		i++;
	clamp_int(&(P_INT(p)), 0, i);
}

/* the initialisation proper */

void init_ptable(void)
{
/* first the trees */
	add_par_tree(PAR_STAR_DET, PAR_NULL,
		     "stars", "Star Detection and Search Options");
	add_par_tree(PAR_FITS_FIELDS, PAR_NULL,
		     "fits", "Fits Field Names");
	add_par_tree(PAR_STAR_DISPLAY, PAR_NULL,
		     "display", "Star Display Options");
	add_par_tree(PAR_STAR_COLORS, PAR_STAR_DISPLAY,
		     "color", "Star Display Colors");
	add_par_tree(PAR_STAR_SHAPES, PAR_STAR_DISPLAY,
		     "shape", "Star Display Shapes");
	add_par_tree(PAR_WCS_OPTIONS, PAR_NULL,
		     "wcs", "Wcs Fitting Options");
	add_par_tree(PAR_OBS_DEFAULTS, PAR_NULL,
		     "obs", "General Observation Setup Data");
	add_par_tree(PAR_FILES, PAR_NULL,
		     "file", "File and device options");
	add_par_tree(PAR_APHOT, PAR_NULL,
		     "aphot", "Aperture photometry defaults");
/* now leaves */
/* leaves for stardet */
	add_par_double(SD_SNR, PAR_STAR_DET, PREC_1, "det_snr", 
		       "SNR for Star Detection (sigmas)", 9.0);
	add_par_int(SD_MAX_STARS, PAR_STAR_DET, 0, "det_maxstars", 
		    "Maximum Number of Detected Stars", 200);
	add_par_int(SD_GSC_MAX_STARS, PAR_STAR_DET, 0, "cat_maxstars", 
		    "Maximum Number of Catalog Stars", 300);
	add_par_double(SD_GSC_MAX_MAG, PAR_STAR_DET, PREC_1, "cat_maxmag", 
		       "Faintest Magnitude for Catalog Stars", 16.0);
	add_par_double(SD_GSC_MAX_RADIUS, PAR_STAR_DET, PREC_1, "cat_maxradius", 
		       "Maximum Radius for Catalog Stars (arcmin)", 120.0);

/* fits fields */
	add_par_string(FN_CRPIX1, PAR_FITS_FIELDS, 0, "crpix1", 
		       "Fits Field for X Coordinate of Reference Pixel", "CRPIX1");
	add_par_string(FN_CRPIX2, PAR_FITS_FIELDS, 0, "crpix2", 
		       "Fits Field for Y Coordinate of Reference Pixel", "CRPIX2");
	add_par_string(FN_CRVAL1, PAR_FITS_FIELDS, 0, "crval1", 
		       "Fits Field for WCS R.A. of Image Center", "CRVAL1");
	add_par_string(FN_CRVAL2, PAR_FITS_FIELDS, 0, "crval2", 
		       "Fits Field for WCS Dec of Image Center", "CRVAL2");
	add_par_string(FN_CDELT1, PAR_FITS_FIELDS, 0, "cdelt1", 
		       "Fits Field for Degrees/Pixel in R.A.", "CDELT1");
	add_par_string(FN_CDELT2, PAR_FITS_FIELDS, 0, "cdelt2", 
		       "Fits Field for Degrees/Pixel in Dec", "CDELT2");
	add_par_string(FN_CROTA1, PAR_FITS_FIELDS, 0, "crota1", 
		       "Fits Field for Field Rotation", "CROTA1");
	add_par_string(FN_EQUINOX, PAR_FITS_FIELDS, 0, "equinox", 
		       "Fits Field for Equinox of WCS", "EQUINOX");
	add_par_string(FN_OBJECT, PAR_FITS_FIELDS, 0, "object", 
		       "Fits Field for Object Name", "OBJECT");
	add_par_string(FN_OBJCTRA, PAR_FITS_FIELDS, 0, "objctra", 
		       "Fits Field for Object R.A.", "OBJCTRA");
	add_par_string(FN_OBJCTDEC, PAR_FITS_FIELDS, 0, "objctdec", 
		       "Fits Field for Object Dec", "OBJCTDEC");
	add_par_string(FN_FILTER, PAR_FITS_FIELDS, 0, "filter", 
		       "Fits Field for Filter Designation", "FILTER");
	add_par_string(FN_EXPTIME, PAR_FITS_FIELDS, 0, "exptime", 
		       "Fits Field for Exposure Time", "EXPTIME");
	add_par_string(FN_JDATE, PAR_FITS_FIELDS, 0, "jdate", 
		       "Fits Field for Julian Date of Observation", "JDATE");
	add_par_string(FN_MJD, PAR_FITS_FIELDS, 0, "mjd", 
		       "Fits Field for Modified Julian Date of Observation", "MJD");
	add_par_string(FN_DATE_OBS, PAR_FITS_FIELDS, 0, "dateobs", 
		       "Fits Field for Date/Time of Observation", "DATE-OBS");
	add_par_string(FN_TELESCOP, PAR_FITS_FIELDS, 0, "telescop", 
		       "Fits Field for Telescope name", "TELESCOP");
	add_par_string(FN_FOCUS, PAR_FITS_FIELDS, 0, "focus", 
		       "Fits Field for Focus Designation", "FOCUS");
	add_par_string(FN_APERTURE, PAR_FITS_FIELDS, 0, "aperture", 
		       "Fits Field for Telescope Aperture", "APERT");
	add_par_string(FN_FLEN, PAR_FITS_FIELDS, 0, "flen", 
		       "Fits Field for Focal Length", "FLEN");
	add_par_string(FN_INSTRUME, PAR_FITS_FIELDS, 0, "instrument", 
		       "Fits Field for Instrument name", "INSTRUME");
	add_par_string(FN_SECPIX, PAR_FITS_FIELDS, 0, "secpix", 
		       "Fits Field for Arcseconds/pixel", "SECPIX");
	add_par_string(FN_OBSERVER, PAR_FITS_FIELDS, 0, "observer", 
		       "Fits Field for Observer name", "OBSERVER");
	add_par_string(FN_LATITUDE, PAR_FITS_FIELDS, 0, "latitude", 
		       "Fits Field for Latitude of observing site", "LAT-OBS");
	add_par_string(FN_LONGITUDE, PAR_FITS_FIELDS, 0, "longitude", 
		       "Fits Field for Longitude of observing site", "LONG-OBS");
	add_par_string(FN_ALTITUDE, PAR_FITS_FIELDS, 0, "altitude", 
		       "Fits Field for Altitude of observing site", "ALT-OBS");
	add_par_string(FN_AIRMASS, PAR_FITS_FIELDS, 0, "airmass", 
		       "Fits Field for Airmas", "AIRMASS");
	add_par_string(FN_SNSTEMP, PAR_FITS_FIELDS, 0, "snstemp", 
		       "Fits Field for Sensor temperature", "SNS_TEMP");
	add_par_string(FN_BINX, PAR_FITS_FIELDS, 0, "binx", 
		       "Fits Field for Horisontal Binning", "CCDBIN1");
	add_par_string(FN_BINY, PAR_FITS_FIELDS, 0, "biny", 
		       "Fits Field for Vertical Binning", "CCDBIN2");
	add_par_string(FN_SKIPX, PAR_FITS_FIELDS, 0, "skipx", 
		       "Fits Field for Horisontal Window Origin", "CCDSKIP1");
	add_par_string(FN_SKIPY, PAR_FITS_FIELDS, 0, "skipy", 
		       "Fits Field for Vertical Window Origin", "CCDSKIP2");
	add_par_string(FN_ELADU, PAR_FITS_FIELDS, 0, "eladu", 
		       "Fits Field for Electrons per ADU", "ELADU");
	add_par_string(FN_DCBIAS, PAR_FITS_FIELDS, 0, "dcbias", 
		       "Fits Field for Average Bias of Frame", "DCBIAS");
	add_par_string(FN_RDNOISE, PAR_FITS_FIELDS, 0, "rdnoise", 
		       "Fits Field for Readout Noise", "RDNOISE");
	add_par_string(FN_FLNOISE, PAR_FITS_FIELDS, 0, "flnoise", 
		       "Fits Field for Multiplicative Noise Coefficient", "FLNOISE");



/* star display colors/sizes */
	add_par_int(SDISP_SIMPLE_SHAPE, PAR_STAR_SHAPES, 0, "simple", 
		    "Shape of Autodected Stars", STAR_SHAPE_SQUARE);
	set_par_choices(SDISP_SIMPLE_SHAPE, star_shapes);
	add_par_int(SDISP_SIMPLE_COLOR, PAR_STAR_COLORS, FMT_COLOR, "simple", 
		    "Color of Autodected Stars", 0x00c000);

	add_par_int(SDISP_SREF_SHAPE, PAR_STAR_SHAPES, 0, "field", 
		    "Shape of Field Stars", STAR_SHAPE_DIAMOND);
	set_par_choices(SDISP_SREF_SHAPE, star_shapes);
	add_par_int(SDISP_SREF_COLOR, PAR_STAR_COLORS, FMT_COLOR, "field", 
		    "Color of Field Stars", 0xff0000);

	add_par_int(SDISP_APSTD_SHAPE, PAR_STAR_SHAPES, 0, "apstd", 
		    "Shape of Photometry Standard Stars", STAR_SHAPE_APHOT);
	set_par_choices(SDISP_APSTD_SHAPE, star_shapes);
	add_par_int(SDISP_APSTD_COLOR, PAR_STAR_COLORS, FMT_COLOR, "apstd", 
		    "Color of Photometry Standard Stars", 0xff0000);

	add_par_int(SDISP_APSTAR_SHAPE, PAR_STAR_SHAPES, 0, "target", 
		    "Shape of Photometry Target Stars", STAR_SHAPE_CROSS);
	set_par_choices(SDISP_APSTAR_SHAPE, star_shapes);
	add_par_int(SDISP_APSTAR_COLOR, PAR_STAR_COLORS, FMT_COLOR, "target", 
		    "Color of Photometry Target Stars", 0xff0000);

	add_par_int(SDISP_CAT_SHAPE, PAR_STAR_SHAPES, 0, "catalog", 
		    "Shape of Catalog Objects", STAR_SHAPE_DIAMOND);
	set_par_choices(SDISP_CAT_SHAPE, star_shapes);
	add_par_int(SDISP_CAT_COLOR, PAR_STAR_COLORS, FMT_COLOR, "catalog", 
		    "Color of Catalog Objects", 0xffff00);

	add_par_int(SDISP_USEL_SHAPE, PAR_STAR_SHAPES, 0, "usel", 
		    "Shape of User-Marked Stars", STAR_SHAPE_CIRCLE);
	set_par_choices(SDISP_USEL_SHAPE, star_shapes);
	add_par_int(SDISP_USEL_COLOR, PAR_STAR_COLORS, FMT_COLOR, "usel", 
		    "Color of User-Marked Stars", 0x00ff00);

//	add_par_int(SDISP_CHECK_SHAPE, PAR_STAR_SHAPES, 0, "check_shape", 
//		    "Shape of Check Stars", STAR_SHAPE_SQUARE);
//	set_par_choices(SDISP_CHECK_SHAPE, star_shapes);
//	add_par_int(SDISP_USEL_COLOR, PAR_STAR_COLORS, FMT_COLOR, "usel_color", 
//		    "Color of Check Stars", 0xff8800);

	add_par_int(SDISP_SELECTED_COLOR, PAR_STAR_COLORS, FMT_COLOR, "selected", 
		    "Color of Selected Stars", 0xffffff);

/* star display opts */
	add_par_int(DO_MIN_STAR_SZ, PAR_STAR_DISPLAY, 0, "min_sz", 
		    "Minimum Display Star Size", 2);
	add_par_int(DO_MAX_STAR_SZ, PAR_STAR_DISPLAY, 0, "max_sz", 
		    "Maximum Display Star Size", 12);
	add_par_int(DO_DEFAULT_STAR_SZ, PAR_STAR_DISPLAY, 0, "default_sz", 
		    "Default Display Star Size", 6);
	add_par_double(DO_PIXELS_PER_MAG, PAR_STAR_DISPLAY, PREC_2, "pix_per_mag", 
		       "Display Pixels per Magnitude", 1.4);
	add_par_double(DO_MAXMAG, PAR_STAR_DISPLAY, PREC_1, "maxmag", 
		       "Highest magnitude for which we increase symbols", 8);
	add_par_double(DO_STAR_SZ_FWHM_FACTOR, PAR_STAR_DISPLAY, PREC_1, "fwhm_sz_fact", 
		       "Size of Star Marker in FWHMs", 3.0);

/* wcs options */

	add_par_double(FIT_SCALE_TOL, PAR_WCS_OPTIONS, PREC_2, "fit_scale", 
		       "Max WCS Scale Change When Auto-Matching Stars", 0.1);
	add_par_double(FIT_MATCH_TOL, PAR_WCS_OPTIONS, PREC_1, "fit_tol", 
		       "Star Position Tolerance When Auto-Matching Stars", 1.5);
	add_par_double(FIT_ROT_TOL, PAR_WCS_OPTIONS, PREC_1, "fit_rot", 
		       "Rotation Tolerance (degrees) for Auto-Matching", 180.0);
	add_par_int(FIT_MIN_PAIRS, PAR_WCS_OPTIONS, 0, "fit_min_pairs", 
		    "Pairs Needed to end Auto-Match", 5);
	add_par_int(FIT_MAX_SKIP, PAR_WCS_OPTIONS, 0, "max_skip", 
		    "Max # of Field Stars Skipped When Searching", 20);
	add_par_int(FIT_MAX_PAIRS, PAR_WCS_OPTIONS, 0, "max_pairs", 
		    "Stop After This Many Matches", 22);
	add_par_double(FIT_MIN_AB_DIST, PAR_WCS_OPTIONS, 0, "min_ab_dist", 
		       "Min Distance Between 'a' and 'b' Stars ", 100.0);

	add_par_double(WCS_ERR_VALIDATE, PAR_WCS_OPTIONS, PREC_2, "validate_err", 
		       "Max Fitted Error for a Valid WCS (pixels)", 1.5);
	add_par_int(WCS_PAIRS_VALIDATE, PAR_WCS_OPTIONS, 0, "validate_pairs", 
		    "Min Fitted Pairs for a Valid WCS", 3);
	add_par_double(WCS_MAX_SCALE_CHANGE, PAR_WCS_OPTIONS, PREC_4, "max_s_chg", 
		       "Max Scale Change per Iteration", 0.02);

	add_par_double(MAX_POINTING_ERR, PAR_WCS_OPTIONS, PREC_2, "max_point_err", 
		       "Max error we accept in telescope pointing (degrees)", 0.05);
	add_par_double(WCS_SEC_PER_PIX, PAR_WCS_OPTIONS, PREC_2, "default_scale", 
		       "Default image scale for wcs fitting (arcsec/pixel)", 1.5);


/* obs */
	add_par_double(OBS_LATITUDE, PAR_OBS_DEFAULTS, FMT_DEC, "latitude", 
		       "Latitude of Observing Site", 44.430556);
	add_par_double(OBS_LONGITUDE, PAR_OBS_DEFAULTS, FMT_DEC, "longitude", 
		       "Longitude of Observing Site", -26.113888);
	add_par_double(OBS_ALTITUDE, PAR_OBS_DEFAULTS, PREC_1, "altitude", 
		       "Altitude of Observing Site (m)", 75);
	add_par_string(OBS_OBSERVER, PAR_OBS_DEFAULTS, 0, "observer", 
		       "Observer Name", "R. Corlan");
	add_par_string(OBS_TELESCOP, PAR_OBS_DEFAULTS, PREC_1, "telescope", 
		       "Telescope Name", "SCT");
	add_par_double(OBS_FLEN, PAR_OBS_DEFAULTS, PREC_1, "flen", 
		       "Telescope Focal Length (cm)", 200);
	add_par_double(OBS_APERTURE, PAR_OBS_DEFAULTS, PREC_1, "aperture", 
		       "Telescope Aperture (cm)", 30);
	add_par_string(OBS_FOCUS, PAR_OBS_DEFAULTS, PREC_1, "focus", 
		       "Focus Designation", "Cassergrain");
	add_par_string(OBS_FILTER_LIST, PAR_OBS_DEFAULTS, PREC_1, "filters", 
		       "List of Filters", "v r b i RI GI BI clear");

	add_par_string(FILE_SCOPE_SERIAL, PAR_FILES, 0, "scope_serial", 
		       "Serial line for telescope control", "/dev/ttyS0");
	add_par_string(FILE_GPS_SERIAL, PAR_FILES, 0, "gps_serial", 
		       "Serial line for gps receiver control", "/dev/ttyS0");
	add_par_string(FILE_UNCOMPRESS, PAR_FILES, 0, "uncompress", 
		       "Command for uncompressing files (like zcat)", "zcat");
	add_par_string(FILE_COMPRESS, PAR_FILES, 0, "compress", 
		       "Command to compress file", "gzip");
	add_par_string(MONO_FONT, PAR_FILES, 0, "mono_font", 
		       "XLFD for the default monospaced font", 
		       "-misc-fixed-medium-r-*-*-*-120-*-*-*-*-*-*");
	add_par_string(FILE_PHOT_OUT, PAR_FILES, 0, "phot_out_fn", 
		       "Default photometry report output file", "phot.out");
	add_par_string(FILE_OBS_PATH, PAR_FILES, 0, "obs_path", 
		       "Path to obs files", ".:../obs");
	add_par_string(FILE_RCP_PATH, PAR_FILES, 0, "rcp_path", 
		       "Path to rcp files", ".:../rcp");
	add_par_string(FILE_GSC_PATH, PAR_FILES, 0, "gsc_path", 
		       "Path to gsc", "");
	add_par_string(FILE_TYCHO2_PATH, PAR_FILES, 0, "tycho2_path", 
		       "Path to tycho2", "");
#define TAB_FORMAT_DEF "name ra dec smag 'v' serr 'v' residual flags collist tablehead"
	add_par_string(FILE_TAB_FORMAT, PAR_FILES, 0, "tab_format", 
		       "Table output format", TAB_FORMAT_DEF);

/* aphot */
	add_par_double(AP_R1, PAR_APHOT, PREC_1, "r1", 
		       "Radius of central aperture", 6);
	add_par_double(AP_R2, PAR_APHOT, PREC_1, "r2", 
		       "Inner radius of sky annulus ", 9);
	add_par_double(AP_R3, PAR_APHOT, PREC_1, "r3", 
		       "Outer radius of sky annulus ", 13);
	add_par_int(AP_AUTO_CENTER, PAR_APHOT, FMT_BOOL, "auto_center", 
		    "Try to center stars before measuring", 1);
	add_par_double(AP_MAX_CENTER_ERR, PAR_APHOT, PREC_2, "center_err", 
		       "Maximum error accepted when centering stars (pixels) ", 2.0);
	add_par_int(AP_SKY_METHOD, PAR_APHOT, 0, "sky_method", 
		    "Method for determining sky value", 0);
	set_par_choices(AP_SKY_METHOD, sky_methods);
	add_par_double(AP_SIGMAS, PAR_APHOT, PREC_1, "sigmas", 
		       "Acceptance band for k-s, mm and smode (sigmas)", 3.0);
	add_par_int(AP_MAX_ITER, PAR_APHOT, 0, "max_iter", 
		       "Max number of iterations for k-s and smode", 20);
	add_par_double(AP_SATURATION, PAR_APHOT, 0, "sat_limit", 
		       "Saturation limit", 28000.0);
	add_par_string(AP_IBAND_NAME, PAR_APHOT, 0, "iband", 
		       "Name of instrumental band used for reduction", 
		       "v");
	add_par_string(AP_SBAND_NAME, PAR_APHOT, 0, "sband", 
		       "Name of standard band used for reduction", 
		       "v(aavso)");
	add_par_string(AP_COLOR_NAME, PAR_APHOT, 0, "color", 
		       "Name of standard color index used for reduction", 
		       "f(gsc2)-j(gsc2)");
	add_par_int(AP_USE_STD_COLOR, PAR_APHOT, 0, "transform_color",
		       "Type of color index used for transforms", 0); 
	set_par_choices(AP_USE_STD_COLOR, color_types);
	add_par_int(AP_FIT_COLTERM, PAR_APHOT, FMT_BOOL, "fit_colterm",
		       "Attempt to fit the color coefficient", 0); 
	add_par_double(AP_COLTERM, PAR_APHOT, 0, "colterm",
		       "Color coefficient of the transform (when not fitted)", 0.0); 
	add_par_double(AP_COLTERM_ERR, PAR_APHOT, 0, "colterm_err",
		       "Relative error of color term", 0.0); 
	add_par_double(AP_DEFAULT_STD_ERROR, PAR_APHOT, 0, "def_std_err",
		       "Default error for catalog magnitudes", 0.05); 
	add_par_double(AP_ALPHA, PAR_APHOT, 0, "alpha",
		       "Alpha robust fitting parameter", 2.0); 
	add_par_double(AP_BETA, PAR_APHOT, 0, "beta",
		       "Beta robust fitting parameter", 2.0); 

}


/*
 * return a pointer to a 1-char string describing the current
 * parameter status. The string is actually a constant
 */
char *status_string(GcxPar p)
{
	if (PAR(p)->flags & PAR_USER) {
		return "[ User Set ]";
	} else if (PAR(p)->flags & PAR_FROM_RC) {
		return "[ From File ]";
	} else {
		return "[ Default ]";
	}
}

/* print a item value for display) */
void make_value_string(GcxPar p, char *c)
{
	switch(PAR_TYPE(p)) {
	case PAR_INTEGER:
		sprintf(c, "%d", P_INT(p));
		break;
	case PAR_DOUBLE:
		switch(PAR(p)->flags & PAR_PREC_FM) {
		case PREC_1:
			sprintf(c, "%.1f", P_DBL(p));
			break;
		case PREC_2:
			sprintf(c, "%.2f", P_DBL(p));
			break;
		case PREC_4:
			sprintf(c, "%.4f", P_DBL(p));
			break;
		case PREC_8:
			sprintf(c, "%.8f", P_DBL(p));
			break;
		default:
			sprintf(c, "%.3f", P_DBL(p));
			break;
		}
		break;
	case PAR_STRING:
		sprintf(c, "%s", P_STR(p));
		break;
	default:
		err_printf("unknown param type %x\n", PAR(p)->flags);
	}
}

/* try to parse the string to a double according to the format
 * return 0 if successfull. 
 * We need to add stuff for formats other than the 'natural'
 */
int try_parse_double(GcxPar p, char *text)
{
	float v;
	int ret;
	ret = sscanf(text, "%f", &v);
	if (ret > 0) {
		P_DBL(p) = v;
		return 0;
	} else {
		return -1;
	}
}

/* try to parse the string to an int according to the format
 * return 0 if successfull. 
 * We need to add stuff for formats other than the 'natural'
 */
int try_parse_int(GcxPar p, char *text)
{
	int v;
	int ret;
	char ** c;

	if (PAR_FORMAT(p) == FMT_OPTION && PAR(p)->choices != NULL) {
		c = PAR(p)->choices;
		v = 0;
		while (*c != NULL) {
			if (!strcmp(text, *c)) {
				P_INT(p) = v;
				return 0;
			}
			v++;
			c++;
		}
		err_printf("try_parse_int: unknown choice %s\n", text);
		return -1;
	} 

	ret = sscanf(text, "%d", &v);
	if (ret > 0) {
		P_INT(p) = v;
		return 0;
	} else {
		return -1;
	}
}

/* change the parameter's value string using a copy of the supplied 
 * text */ 
void change_par_string(GcxPar p, char *text)
{
	if (P_STR(p) != NULL) {
		free(P_STR(p));
		P_STR(p) = NULL;
	}
	P_STR(p) = malloc(strlen(text)+1);
	if (P_STR(p) != NULL)
		strcpy(P_STR(p), text);
}

/* change the parameter's value string using a copy of the supplied 
 * text */ 
void change_par_default_string(GcxPar p, char *text)
{
	if (PAR(p)->defval.s != NULL) {
		free(PAR(p)->defval.s);
		PAR(p)->defval.s = NULL;
	}
	PAR(p)->defval.s = malloc(strlen(text)+1);
	if (PAR(p)->defval.s != NULL)
		strcpy(PAR(p)->defval.s, text);
}


/* parse the string according to the param's type and format
 * and update it's value. return 0 if the parse was successfull,
 * or a negative error
 */
int try_update_par_value(GcxPar p, char *text)
{
	switch(PAR_TYPE(p)) {
	case PAR_INTEGER:
		return try_parse_int(p, text);
	case PAR_DOUBLE:
		return try_parse_double(p, text);
	case PAR_STRING:
		change_par_string(p, text);
		return 0;
	default:
		err_printf("unknown param type %x\n", PAR(p)->flags);
		return -1;
	}
}

/* restore an item's default (recurses through children) */
/* the status of ancestor trees is not changed */
void par_item_restore_default(GcxPar p)
{
	if (PAR(p)->flags & (PAR_TREE)) { // recurse tree
		GcxPar pp;
		PAR(p)->flags &= ~PAR_USER;
		pp = PAR(p)->child;
		while(pp != PAR_NULL) {
			par_item_restore_default(pp);
			pp = PAR(pp)->next;
		}
	} else if (PAR(p)->flags & (PAR_USER)) {
			if (PAR_TYPE(p) == PAR_STRING) {
				change_par_string(p, PAR(p)->defval.s);
			} else {
				memcpy(&(PAR(p)->val), &(PAR(p)->defval), sizeof(union pval));
			}
			PAR(p)->flags &= ~PAR_USER; 
			PAR(p)->flags &= ~PAR_TO_SAVE;
	}
}

/* return the logic or of all children of a par subtree */
int or_child_type(GcxPar p, int type)
{
	if (PAR_TYPE(p) != PAR_TREE || PAR(p)->child == PAR_NULL) 
		return (type | PAR(p)->flags);
	else {
		type |= PAR(p)->flags;
		p = PAR(p)->child;
		while(p != PAR_NULL) {
			type |= or_child_type(p, type);
			p = PAR(p)->next;
		}
	}
	return type;
}

/* update the state of all ancestors to a node */
void update_ancestors_state(GcxPar p)
{
	p = PAR(p)->parent;
	while (p != PAR_NULL && PAR_TYPE(p) == PAR_TREE) {
		int type = 0;
		PAR(p)->flags &= ~(PAR_USER | PAR_FROM_RC);
		type = or_child_type(p, 0);
		PAR(p)->flags |= (type & (PAR_USER | PAR_FROM_RC));
		p = PAR(p)->parent;
	}
}
/* print indents*/
//static void fprint_indent(FILE *fp, int level)
//{
//	if (level > 0)
//		fputc('\t', fp);
//}

/* print a leaf node */
static void fprint_par_leaf(FILE *fp, GcxPar p, char *path)
{
	char **c;
	int i;
	char buf[256];
//	fprint_indent(fp, level);
	fprintf(fp, "# %s\n", PAR(p)->comment);
	if (PAR_TYPE(p) == PAR_INTEGER 
	    && PAR_FORMAT(p) == FMT_OPTION 
	    && PAR(p)->choices != NULL) {
//		fprint_indent(fp, level);
		c = PAR(p)->choices;
		fprintf(fp, "# Choices: ");
		while (*c != NULL) {
			fprintf(fp, "%s", *c);
			c++;
			if (*c != NULL)
				fprintf(fp, ", ");
		}
		fputc('\n', fp);
		c = PAR(p)->choices;
		for (i = 0; i < P_INT(p) && *c != NULL; i++)
			c++;
//		fprint_indent(fp, level);
		if ((PAR_FLAGS(p) & (PAR_FROM_RC | PAR_USER)) == 0) {
			fprintf(fp, "# ");
		}
		fprintf(fp, "%s.%s %s\n\n", path, PAR(p)->name, *c);
	} else {
		make_value_string(p, buf);
//		fprint_indent(fp, level);
		if ((PAR_FLAGS(p) & (PAR_FROM_RC | PAR_USER)) == 0) {
			fprintf(fp, "# ");
		}
		if (PAR_TYPE(p) == PAR_STRING) {
			fprintf(fp, "%s.%s '%s'\n\n", path, PAR(p)->name, buf);
		} else {
			fprintf(fp, "%s.%s %s\n\n", path, PAR(p)->name, buf);
		}
	}
}

static void fprint_item(FILE *fp, GcxPar p, char *path)
{
	char npath[256];
	int pp = p;

//	d3_printf("fprint_item: path: %s p= %d\n", path, p);

	if(PAR_TYPE(p) == PAR_TREE) {
		sprintf(npath, "%s.%s", path, PAR(p)->name);
		fprintf(fp, "#### %s\n#\n", PAR(p)->comment);
		p = PAR(pp)->child;
/* output the leaves first */
		while (p != PAR_NULL) {
			if (PAR_TYPE(p) != PAR_TREE) {
				fprint_item(fp, p, npath);
			}
			p = PAR(p)->next;
		}
/* then the subtrees */
		p = PAR(pp)->child;
		while (p != PAR_NULL) {
			if (PAR_TYPE(p) == PAR_TREE)
				fprint_item(fp, p, npath);
			p = PAR(p)->next;
		}
	} else {
//		while (*path != 0) {
//			if (*path ++ == '.')
//				dots++;
//		}
//		d3_printf("leaf path is %s\n", path);
		fprint_par_leaf(fp, p, path);
	}
}



/* save params in a file 
 * if p is PAR_NULL, save all the params, otherwise just the 
 * given param and it's children
 * returns a negative error code, or 0 for success
 */
void fprint_params(FILE *fp, GcxPar p)
{
	if (p != PAR_NULL) {

/* search for the proper path */
		fprint_item(fp, p, ".");
	} else {
		p = PAR_FIRST;
		while (p != PAR_NULL) {
			fprint_item(fp, p, "");
			p = PAR(p)->next;
		}
	}
}

/* definitions for the lexical analyser */
typedef enum {
	T_START,
	T_STRING,
	T_NUMBER,
	T_WORD,
} ParLexState;
static inline int is_string_delim(char c) {return (c == '\'') || (c=='"');}
static inline int is_name_char(char c) {return (isalnum(c)) || (c=='_');}
static inline int is_number_char(char c) {return (isdigit(c) || (c) == '.' ||
					      (c) == 'e' || (c) == 'E'); }
static inline int is_number_start(char c) {return (isdigit(c) || c == '-'); }
/* simple lexical scanner for param file reading 
 * returns the token type 
 * updates the start and end pointers, which 
 * give the first char of the token value, 
 * and the char immediately following the value;
 * Also updates the text pointer to
 * the next char the reader will look at. 
 * The scanner never goes back, so the text behind pos
 * can be altered at will.
 * when called with text=NULL, it resets the lex state
 */
int next_token(char **text, char **start, char **end)
{
	static ParLexState state = T_START;
	static char string_delim;
	char c;

	if (text == NULL) {
		state = T_START;
		return TOK_EOL;
	}

	while (**text != 0) {
		c = **text; 
		(*text) ++;
		switch(state) {
		case T_START:
			if (isspace(c))
				break;
			if (c == '.') {
				*start = *text - 1;
				*end = *text;
				return TOK_DOT;
			}
			if (c == '#')
				return TOK_EOL;
			if (is_string_delim(c)) {
				string_delim = c;
				*start = *text;
				*end = *text;
				state = T_STRING;
				break;
			}
			if (is_number_start(c)) {
				*start = *text - 1;
				*end = *text;
				state = T_NUMBER;
				break;
			}
			if (is_name_char(c)) {
				*start = *text - 1;
				*end = *text;
				state = T_WORD;
				break;
			}
			*start = *text - 1;
			*end = *text;
			return TOK_PUNCT;
		case T_STRING:
			if (c == string_delim) {
				state = T_START;
				return TOK_STRING;
			}
			*end = *text;
			break;
		case T_WORD:
			if (!is_name_char(c)) {
				state = T_START;
				return TOK_WORD;
			}
			*end = *text;
			break;
		case T_NUMBER:
			if (!is_number_char(c)) {
				state = T_START;
				return TOK_NUMBER;
			}
			*end = *text;
			break;
		}
	}
	switch(state) {
	case T_WORD:
		state = T_START;
		return TOK_WORD;
		break;
	case T_NUMBER:
		state = T_START;
		return TOK_NUMBER;
		break;
	default:
		state = T_START;
		return TOK_EOL;
	}
}

/*
 * try to assign the given value string to par p
 * return 0 if sucessfull
 */
static int parse_token_val(GcxPar p, char *start, int flags)
{
	int ret;
//	d3_printf("par %d gets value %s\n", p, start);
	ret = try_update_par_value(p, start);
	if (ret < 0) {
		err_printf("value %s for par %d (%s) doesn't parse\n",
			   start, p, PAR(p)->name);
		return -1;
	} else {
		if (PAR_TYPE(p) == PAR_STRING) {
			change_par_default_string(p, PAR(p)->val.s);
		} else {
			memcpy(&(PAR(p)->defval), &(PAR(p)->val), sizeof(union pval));
		}
		PAR_FLAGS(p) |= flags;
		return 0;
	}
}


/* lookup a node by name within the peers of p */
static GcxPar node_lookup(GcxPar p, char *name, int len)
{
	while (p != PAR_NULL) {
		if (name_matches(PAR(p)->name, name, len))
//		if (PAR(p)->name && !strncmp(name, PAR(p)->name, len)) 
			return p;
		p = PAR(p)->next;
	}
	return PAR_NULL;
}


typedef enum {
	P_START,
	P_NLINE,
	P_DOT,
	P_REL_TREE,
	P_ABS_TREE,
	P_LEAF,
} ParParseStat;
/*
 * scan a branch of the param tree
 * if p is PAR_NULL, the root of the file is placed
 * at the root of the par tree, else it is assumed to be p
 */
int fscan_params(FILE *fp, GcxPar root)
{
	ParParseStat state = P_START;
	char *line=NULL, *text, *start, *end;
	int ret = 0, len, token;
	GcxPar pp = PAR_NULL; /* tree we're reading in */
	GcxPar p = PAR_NULL; /* current item we read in */

	if (root == PAR_NULL) {
		root = PAR_FIRST;
	} else if (PAR(root)->child == PAR_NULL) {
		err_printf("%d - not a tree\n", root);
		return -1;
	} else {
		root = PAR(root)->child;
	}
	pp = p = root;

	while (ret >= 0) {
//		if (p)
//			d3_printf("parse state is %d p=%d (%s)\n", state, p, PAR(p)->name);
//		else 
//			d3_printf("parse state is %d p=%d\n", state, p);
		switch(state) {
		case P_START:
			line = NULL;
			len = 0;
			pp = root;
			p = root;
			ret = 0;
			state = P_NLINE;
			break;
		case P_NLINE:
			ret = getline (&line, &len, fp);
			if (ret < 0) 
				break;
			text = line;
			token = next_token(&text, &start, &end);
			switch(token) {
			case TOK_EOL:
				ret = 0;
				break;
			case TOK_WORD: /* start of a relative path */
				p = node_lookup(pp, start, end-start);
//				d3_printf("tok_word %s\np=%d\n", start, p);
				if (p == PAR_NULL) { 
					*(end)=0;
					err_printf("fscan_params: unknown name %s under node %d\n",
						   start, PAR(pp)->parent);
					state = P_NLINE;
				} else {
					if (PAR_TYPE(p) != PAR_TREE) {
						state = P_LEAF;
					} else {
						p = PAR(p)->child;
						state = P_REL_TREE;
					}
				}
				break;
			case TOK_DOT: /* start of an absolute path */
				pp = root;
				state = P_ABS_TREE;
				break;
			default:
				*(end)=0;
				err_printf("fscan_params: unexpected token %s at start of line\n",
					   start);
				state = P_NLINE;
				break;
			}
			break;
		case P_REL_TREE:
			token = next_token(&text, &start, &end);
			if (token == TOK_DOT)
				break;
			if (token != TOK_WORD) {
				*(end)=0;
				err_printf("fscan_params: unexpected token %s at start of line\n",
					   start);
				state = P_NLINE;
				break;
			}
			p = node_lookup(p, start, end-start);
			if (p == PAR_NULL) { 
				*(end)=0;
				err_printf("fscan_params: unknown name %s \n",
					   start);
				state = P_NLINE;
				break;
			} else {
				if (PAR_TYPE(p) != PAR_TREE) {
					state = P_LEAF;
				} else {
					p = PAR(p)->child;
				}
			}
			break;
		case P_ABS_TREE:
			token = next_token(&text, &start, &end);
			if (token == TOK_DOT)
				break;
			if (token == TOK_EOL) {
				state = P_NLINE;
				break;
			}
			if (token == TOK_WORD) {
				p = node_lookup(pp, start, end-start);
				if (p == PAR_NULL) { 
					*(end)=0;
					err_printf("fscan_params: unknown name %s under node %d\n",
						   start, PAR(pp)->parent);
				} else {
					if (PAR_TYPE(p) != PAR_TREE) {
						state = P_LEAF;
					} else {
						pp = PAR(p)->child;
					}
				}
			} else {
				*(end)=0;
				err_printf("fscan_params: unexpected token %s\n",
					   start);
				state = P_NLINE;
				break;
			}
			break;
		case P_LEAF: 
			token = next_token(&text, &start, &end);
			if (token == TOK_DOT)
				break;
			if (token == TOK_WORD || token == TOK_STRING || token == TOK_NUMBER) {
				*(end) = 0;
				parse_token_val(p, start, PAR_FROM_RC);
				state = P_NLINE;
				break;
			}
			*(end) = 0;
			err_printf("fscan_params: unexpected token %s instead of par value\n",
				   start);
			break;
		default:
			break;
		}
	}
	if (line != NULL)
		free(line);
	return 0;
}

/* set a parameter by it's pathname. line contains the name and value on a single
 * line, separated by spaces; return 0 if the par has been sucessfully set
 * only one parameter per line can be set */
int set_par_by_name(char *line)
{
	ParParseStat state = P_START;
	char *text, *start, *end;
	int ret = 0, len, token;
	GcxPar pp; /* tree we're reading in */
	GcxPar p; /* current item we read in */

	pp = p = PAR_FIRST;
	len = strlen(line);
	ret = 0;
	state = P_NLINE;
	text = line;

	while (ret >= 0) {
		switch(state) {
		case P_NLINE:
			token = next_token(&text, &start, &end);
			switch(token) {
			case TOK_EOL:
				return -1;
			case TOK_WORD: /* start of a relative path */
				p = node_lookup(pp, start, end-start);
//				d3_printf("tok_word %s\np=%d\n", start, p);
				if (p == PAR_NULL) { 
					*(end)=0;
					err_printf("set_par_by_name: unknown name %s "
						   "under node %d\n",
						   start, PAR(pp)->parent);
					return -1;
				} else {
					if (PAR_TYPE(p) != PAR_TREE) {
						state = P_LEAF;
					} else {
						p = PAR(p)->child;
						state = P_REL_TREE;
					}
				}
				break;
			case TOK_DOT: /* start of an absolute path */
				break;
			default:
				*(end)=0;
				err_printf("fscan_params: unexpected token %s at start of line\n",
					   start);
				return -1;
			}
			break;
		case P_REL_TREE:
			token = next_token(&text, &start, &end);
			if (token == TOK_DOT)
				break;
			if (token != TOK_WORD) {
				*(end)=0;
				err_printf("fscan_params: unexpected token %s at start of line\n",
					   start);
				return -1;
			}
			p = node_lookup(p, start, end-start);
			if (p == PAR_NULL) { 
				*(end)=0;
				err_printf("fscan_params: unknown name %s \n",
					   start);
				return -1;
			} else {
				if (PAR_TYPE(p) != PAR_TREE) {
					state = P_LEAF;
				} else {
					p = PAR(p)->child;
				}
			}
			break;
		case P_LEAF: 
			token = next_token(&text, &start, &end);
			if (token == TOK_DOT)
				break;
			if (token == TOK_WORD || token == TOK_STRING || token == TOK_NUMBER) {
				*(end) = 0;
				if (!parse_token_val(p, start, PAR_USER))
					return 0;
				else
					return -1;
			}
			*(end) = 0;
			err_printf("fscan_params: unexpected token %s instead of par value\n",
				   start);
			return -1;
		default:
			return -1;
		}
	}
	return -1;
}


/* check that the name staring at name and len chars long matches the
 * string str. return 1 for a match */
int name_matches(char *str, char *name, int len)
{
//	d3_printf("matches? %s %s %d\n", str, name, len);
	while (*str != 0 && len > 0) {
		if (*str == *name) {
			str ++;
			name ++;
			len --;
		} else {
			return 0;
		}
	}
//	d3_printf("%s\n", (*str || len) ? "no" : "yes");
	return !(*str || len);
}
