/* Functions that handle 'parameters' - global options that are 
 * saved/loaded from rc files
 *
 * parameters are created at compile-time. Their names are built
 * hierarchically, but accessed from the rest of the code
 * trough macros taking constant indices as arguments
 */

#ifndef _PARAMS_H_
#define _PARAMS_H_

union pval {
	int i;
	unsigned int u;
	double d;
	char *s;
};

struct param {
	int flags; /* data type & flags of the parameter */
	char ** choices; /* list of param choices / limits */
	int next; /* this node's next sibling */
	int child; /* this node's first child (NULL for leaf nodes) */
	int parent; /* this node's parent (NULL for the root node) */
	char *name; /* name as visible in the rc file */
	char *comment; /* a comment describing the parameter */
	void * item; /* the item used to edit the parameter
			 * this is only valid while the edit tree is active,
			 * so sould be only accessed from callbacks! */
	union pval val;
	union pval defval;
};

typedef enum {
	TOK_EOL,
	TOK_WORD,
	TOK_DOT,
	TOK_STRING,
	TOK_NUMBER,
	TOK_PUNCT,
} TokenType;


extern struct param ptable[];


/* param types */
#define PAR_TYPE_MASK 0xff
#define PAR_INVALID 0
#define PAR_TREE 1
#define PAR_INTEGER 0x10
#define PAR_DOUBLE 0x30
#define PAR_STRING 0x40

/* the format field tells us how we should print and read the parameter */
#define PAR_FORMAT_FM 0x0f00

#define FMT_NAT 0
#define FMT_RA 0x100 /* dms r.a mode for double */
#define FMT_DEC 0x200 /* dms r.a mode for double */
#define FMT_BOOL 0x300 /* boolean (for int) */
#define FMT_HEX 0x400 /* hex (for int) */
#define FMT_COLOR 0x500 /* color (r,g,b) format */
#define FMT_OPTION 0x600 /* one of several options (for int) 
			  * the options are kept in the NULL-terminated
			  * choices string list, the value being each option's
			  * position */

#define PAR_PREC_FM 0xf000 /* precision field for double */
#define PAR_PREC_FS 12 

#define PREC_1 (1 << PAR_PREC_FS)
#define PREC_2 (2 << PAR_PREC_FS)
#define PREC_4 (4 << PAR_PREC_FS)
#define PREC_8 (8 << PAR_PREC_FS)

/* flags */
#define PAR_FROM_RC 0x10000 /* read from a rc file */
#define PAR_USER 0x20000 /* changed by user */
#define PAR_TO_SAVE 0x40000 /* to be saved */


/* macros to acces the par values */
#define P_INT(p) (ptable[p].val.i) 
#define P_UINT(p) (ptable[p].val.u)
#define P_DBL(p) (ptable[p].val.d)
#define P_STR(p) ((ptable[p].val.s))
#define PAR(p) ((p)== 0 ? NULL : &(ptable[p]))
#define PAR_TYPE(p) (ptable[p].flags & PAR_TYPE_MASK)
#define PAR_FORMAT(p) (ptable[p].flags & PAR_FORMAT_FM)
#define PAR_PREC(p) (ptable[p].flags & PAR_PREC_FM)
#define PAR_FLAGS(p) (ptable[p].flags)
/* parameter indices */


typedef enum {
/* trees */
	PAR_NULL ,
	PAR_STAR_DET , /* star detection */
	PAR_FITS_FIELDS , /* fits field names */
	PAR_STAR_DISPLAY , /* options for star drawing */
	PAR_STAR_COLORS , /* colors/shapes of stars */
	PAR_STAR_SHAPES , /* colors/shapes of stars */
	PAR_WCS_OPTIONS , /* options for wcs fitting */
	PAR_OBS_DEFAULTS , /* defaults for obs fields */
	PAR_FILES , /* defaults for obs fields */
	PAR_APHOT , /* defaults for obs fields */

/* leaves for stardet */
	SD_SNR ,
	SD_MAX_STARS ,

	SD_GSC_MAX_MAG ,
	SD_GSC_MAX_RADIUS ,
	SD_GSC_MAX_STARS ,

/* leaves for fits field names */
	FN_CRVAL1 ,
	FN_CRVAL2 ,
	FN_CDELT1 ,
	FN_CDELT2 ,
	FN_CROTA1 ,
	FN_EQUINOX ,
	FN_OBJECT ,
	FN_OBJCTRA ,
	FN_OBJCTDEC ,
	FN_FILTER ,
	FN_EXPTIME ,
	FN_JDATE ,
	FN_MJD ,
	FN_DATE_OBS ,
	FN_TELESCOP ,
	FN_INSTRUME ,
	FN_OBSERVER ,
	FN_LATITUDE ,
	FN_LONGITUDE ,
	FN_ALTITUDE ,
	FN_AIRMASS ,

/* star display */
	SDISP_SIMPLE_SHAPE ,
	SDISP_SIMPLE_COLOR ,
	SDISP_SREF_SHAPE ,
	SDISP_SREF_COLOR ,
	SDISP_APSTD_SHAPE ,
	SDISP_APSTD_COLOR ,
	SDISP_APSTAR_SHAPE ,
	SDISP_APSTAR_COLOR ,
	SDISP_CAT_SHAPE ,
	SDISP_CAT_COLOR ,
	SDISP_USEL_SHAPE ,
	SDISP_USEL_COLOR ,
	SDISP_CHECK_SHAPE ,
	SDISP_CHECK_COLOR ,
	SDISP_SELECTED_COLOR ,

/* star dopts */
	DO_MIN_STAR_SZ ,
	DO_MAX_STAR_SZ ,
	DO_DEFAULT_STAR_SZ ,
	DO_PIXELS_PER_MAG ,
	DO_STAR_SZ_FWHM_FACTOR ,
	DO_MAXMAG ,

/* wcs opts */
	WCS_ERR_VALIDATE ,
	WCS_PAIRS_VALIDATE ,
	WCS_MAX_SCALE_CHANGE ,
	WCS_SEC_PER_PIX ,

	FIT_SCALE_TOL ,
	FIT_MATCH_TOL ,
	FIT_ROT_TOL ,
	FIT_MIN_PAIRS ,
	FIT_MAX_PAIRS ,
	FIT_MAX_SKIP ,
	FIT_MIN_AB_DIST ,

	MAX_POINTING_ERR ,
/* obs */

	OBS_FLEN ,
	OBS_APERTURE ,
	OBS_FOCUS ,
	OBS_TELESCOP ,
	OBS_FILTER_LIST ,
	OBS_LATITUDE ,
	OBS_LONGITUDE ,
	OBS_ALTITUDE ,
	OBS_OBSERVER ,

/* more fits fields */
	FN_SNSTEMP ,
	FN_FOCUS ,
	FN_APERTURE ,
	FN_BINX ,
	FN_BINY ,
	FN_SKIPX ,
	FN_SKIPY ,
	FN_ELADU ,
	FN_DCBIAS ,
	FN_RDNOISE ,
	FN_FLNOISE ,
	FN_FLEN ,
	FN_CRPIX1 , 
	FN_CRPIX2 ,
	FN_SECPIX ,

	FILE_SCOPE_SERIAL ,
	FILE_GPS_SERIAL ,
	FILE_UNCOMPRESS ,
	FILE_COMPRESS ,
	MONO_FONT ,
	FILE_PHOT_OUT ,
	FILE_RCP_PATH ,
	FILE_OBS_PATH ,
	FILE_GSC_PATH ,
	FILE_TYCHO2_PATH ,
	FILE_TAB_FORMAT ,

	AP_R1 ,
	AP_R2 ,
	AP_R3 ,
	AP_MAX_CENTER_ERR ,
	AP_SKY_METHOD ,
	AP_SIGMAS ,
	AP_MAX_ITER ,
	AP_SATURATION ,
	AP_AUTO_CENTER ,
	AP_IBAND_NAME ,
	AP_SBAND_NAME ,
	AP_COLOR_NAME ,
	AP_DEFAULT_STD_ERROR ,
	AP_ALPHA ,
	AP_BETA ,
	AP_USE_STD_COLOR ,
	AP_FIT_COLTERM,
	AP_COLTERM,
	AP_COLTERM_ERR,
	PAR_TABLE_SIZE
} GcxPar;

#define PAR_FIRST PAR_STAR_DET


/* function prototypes */
/* params.c */
extern void init_ptable(void);
char *status_string(GcxPar p);
void make_value_string(GcxPar p, char *c);
int try_parse_double(GcxPar p, char *text);
int try_parse_int(GcxPar p, char *text);
int try_update_par_value(GcxPar p, char *text);
void change_par_string(GcxPar p, char *text);
int or_child_type(GcxPar p, int type);
int fscan_params(FILE *fp, GcxPar root);
int next_token(char **text, char **start, char **end);
void fprint_params(FILE *fp, GcxPar p);
int name_matches(char *str, char *name, int len);
int set_par_by_name(char *line);

#endif
