/*******************************************************************************
  Copyright(c) 2000 - 2003 Radu Corlan. All rights reserved.
  
  This program is free software; you can redistribute it and/or modify it 
  under the terms of the GNU General Public License as published by the Free 
  Software Foundation; either version 2 of the License, or (at your option) 
  any later version.
  
  This program is distributed in the hope that it will be useful, but WITHOUT 
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
  more details.
  
  You should have received a copy of the GNU General Public License along with
  this program; if not, write to the Free Software Foundation, Inc., 59 
  Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  
  The full GNU General Public License is included in this distribution in the
  file called LICENSE.
  
  Contact Information: radu@corlan.net
*******************************************************************************/

#define _GNU_SOURCE
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <libgen.h>
#include <math.h>

#include "gcx.h"
#include "params.h"
#include "photsolve.h"


static void print_pfit_star(struct pfit_star *pfs)
{
	return;
	d3_printf("smag=%.3f smerr=%.3f imag=%.3f imerr=%.3f col=%.3f cerr=%.3f\n",
		  pfs->smag, pfs->smag_err, pfs->imag, pfs->imag_err, 
		  pfs->color, pfs->color_err);
	d3_printf("            w=%.3f r=%.3f\n", pfs->weight, pfs->residual);
}


/* see if we have data that can be used for a color transform fit */
int can_fit_colors(struct pfit_star *stds, int nstd)
{
	int i;
	double sesq = 0.0, iesq = 0.0, cesq = 0.0;
	double csum=0.0, csumsq=0.0, cs;
 
	if (nstd < 4)
		return 0;

	for (i = 0; i < nstd; i++) {
		sesq += sqr(stds[i].smag_err);
		iesq += sqr(stds[i].imag_err);
		cesq += sqr(stds[i].color_err);
		csum += stds[i].color;
		csumsq += sqr(stds[i].color);
	}
	cs = sqrt((csumsq - csum) / nstd);
	if (cs < 2 * sqrt(sesq/nstd))
		return 0;
	if (cs < 2 * sqrt(iesq/nstd))
		return 0;
	if (cs < 2 * sqrt(cesq/nstd))
		return 0;
	return 1;
}

/* weighting function suggested by Stetson */
static double weight_function(struct pfit_star *std)
{
	double err, w;

	if (std->smag == 0.0 && std->smag_err == 0.0)
		return 0.000000001;
	if ((std->flags & PFIT_SMAG_VALID) == 0)
		return 0.000000001;

	err = sqr(std->imag_err) + sqr(std->smag_err);
	if (err == 0)
		return 1.0;
	w = 1.0 / (1.0 + pow(abs(std->residual)/(P_DBL(AP_ALPHA) * sqrt(err)), P_DBL(AP_BETA)));
	w /= err;
	return w;
}


/* return the residual of the star after the color transform 
 * the zero_point transform is not applied! */
static double std_residual(struct pfit_star *stds, struct star_transform *trans)
{
	double residual;

	residual = stds->smag - stds->imag;
	if ((trans->flags & TRANSFORM_COLTERM_VALID) 
	    && (stds->flags & PFIT_COLOR_VALID))
		residual += stds->color * trans->colterm;
	return residual;
}

static void fit_std_no_colors(struct pfit_star *stds, int nstd, 
			      struct star_transform *trans, double *meu)
{
	int i, it=0;
	double e, m;
	double rm=0, rmo=0, w;
	int stdc=0;
	double maxw = 0.0;

	trans->flags &= ~TRANSFORM_ZEROP_VALID;

	for (i = 0; i < nstd; i++) {
		stds[i].residual = 0;
		stds[i].weight = weight_function(stds+i);
		if (stds[i].flags & PFIT_SMAG_VALID)
			stdc ++;
	}
	if (stdc == 0)
		return;

	do {
		rm = 0, w=0;
		it++;
		for (i = 0; i < nstd; i++) {
			rm += (std_residual(stds+i, trans) - rmo) * stds[i].weight;
			w += stds[i].weight;
		}
		if (abs(rm) < 0.0001)
			break;
		rmo += rm / w;
		for (i = 0; i < nstd; i++) {
			stds[i].residual = std_residual(stds+i, trans) - rmo;
			stds[i].weight = weight_function(stds+i);
		}
	} while (it < 50);
	e=0;
	for (i = 0; i < nstd; i++) {
//		stds[i].imag += rmo;
		stds[i].residual = std_residual(stds+i, trans) - rmo;
		e += sqr(stds[i].residual) * stds[i].weight;
		if (stds[i].weight > maxw)
			maxw = stds[i].weight;
	}
	m = sqrt(e / (nstd - 1));
	trans->zerop = rmo;
	trans->cov_zz = e / w;
	if (meu != NULL)
		*meu = m;
	trans->flags |= TRANSFORM_ZEROP_VALID;
	for (i = 0; i < nstd; i++) {
		stds[i].weight /= maxw;
	}
//	d3_printf("%d iterations, e=%.3f, m=%.3f\n", it, e, m);
}

static void fit_std_colors(struct pfit_star *stds, int nstd, struct star_transform *trans,
			   double *meu)
{
	err_printf("colors_fit: not implemented\n");
	fit_std_no_colors(stds, nstd, trans, meu);
}

void fit_std_stars(struct pfit_star *stds, int nstd, struct star_transform *trans,
		   double *meu)
{
	int i;
/* check if we should try to fit the color transform */
	if ((!(trans->flags & TRANSFORM_COLTERM_VALID)) 
	    && can_fit_colors(stds, nstd)) {
		fit_std_colors(stds, nstd, trans, meu);
	} else {
		fit_std_no_colors(stds, nstd, trans, meu);
	}

	for (i=0; i<nstd; i++) {
		print_pfit_star(stds+i);
	}
}

void transform_stars(struct pfit_star *pgms, int npgm, struct star_transform *trans)
{
	int i;
	double y;
	double y_err;

	if (trans->flags & TRANSFORM_ZEROP_VALID) {
		for(i = 0; i < npgm; i++) {
			if ((pgms[i].flags & PFIT_COLOR_VALID) 
			    && (trans->flags & TRANSFORM_ZEROP_VALID)) {
				gsl_fit_linear_est(pgms[i].color, 
						   trans->zerop, trans->colterm,
						   trans->cov_zz, trans->cov_zc, trans->cov_cc,
						   &y, &y_err);
				pgms[i].smag = y + pgms[i].imag;
				pgms[i].smag_err = sqrt(sqr(y_err) + sqr(pgms[i].imag_err));
				pgms[i].flags |= PFIT_SMAG_VALID | PFIT_SMAGERR_VALID;
			} else {
				gsl_fit_linear_est(0, 
						   trans->zerop, 0,
						   trans->cov_zz, 0, 0,
						   &y, &y_err);
				pgms[i].smag = y + pgms[i].imag;
				pgms[i].smag_err = sqrt(sqr(y_err) + sqr(pgms[i].imag_err));
				pgms[i].flags |= PFIT_SMAG_VALID | PFIT_SMAGERR_VALID;
			}
//		d3_printf("zerop = %.3f\n", trans->zerop);
//		pgms[i].imag += trans->zerop;
		}
	}
				   
	for (i=0; i<npgm; i++) {
		print_pfit_star(pgms+i);
	}
}


/* taken from gsl-1.3 */

/* fit/linear.c
 * 
 * Copyright (C) 2000 Brian Gough
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#define GSL_SUCCESS 0
/* Fit the data (x_i, y_i) to the linear relationship 

   Y = c0 + c1 x

   returning, 

   c0, c1  --  coefficients
   cov00, cov01, cov11  --  variance-covariance matrix of c0 and c1,
   sumsq   --   sum of squares of residuals 

   This fit can be used in the case where the errors for the data are
   uknown, but assumed equal for all points. The resulting
   variance-covariance matrix estimates the error in the coefficients
   from the observed variance of the points around the best fit line.
*/

int
gsl_fit_linear (const double *x, const size_t xstride,
		const double *y, const size_t ystride,
		const size_t n,
		double *c0, double *c1,
		double *cov_00, double *cov_01, double *cov_11, double *sumsq)
{
  double m_x = 0, m_y = 0, m_dx2 = 0, m_dxdy = 0;

  size_t i;

  for (i = 0; i < n; i++)
    {
      m_x += (x[i * xstride] - m_x) / (i + 1.0);
      m_y += (y[i * ystride] - m_y) / (i + 1.0);
    }

  for (i = 0; i < n; i++)
    {
      const double dx = x[i * xstride] - m_x;
      const double dy = y[i * ystride] - m_y;

      m_dx2 += (dx * dx - m_dx2) / (i + 1.0);
      m_dxdy += (dx * dy - m_dxdy) / (i + 1.0);
    }

  /* In terms of y = a + b x */

  {
    double s2 = 0, d2 = 0;
    double b = m_dxdy / m_dx2;
    double a = m_y - m_x * b;

    *c0 = a;
    *c1 = b;

    /* Compute chi^2 = \sum (y_i - (a + b * x_i))^2 */

    for (i = 0; i < n; i++)
      {
	const double dx = x[i * xstride] - m_x;
	const double dy = y[i * ystride] - m_y;
	const double d = dy - b * dx;
	d2 += d * d;
      }

    s2 = d2 / (n - 2.0);	/* chisq per degree of freedom */

    *cov_00 = s2 * (1.0 / n) * (1 + m_x * m_x / m_dx2);
    *cov_11 = s2 * 1.0 / (n * m_dx2);

    *cov_01 = s2 * (-m_x) / (n * m_dx2);

    *sumsq = d2;
  }

  return GSL_SUCCESS;
}


/* Fit the weighted data (x_i, w_i, y_i) to the linear relationship 

   Y = c0 + c1 x

   returning, 

   c0, c1  --  coefficients
   s0, s1  --  the standard deviations of c0 and c1,
   r       --  the correlation coefficient between c0 and c1,
   chisq   --  weighted sum of squares of residuals */

int
gsl_fit_wlinear (const double *x, const size_t xstride,
		 const double *w, const size_t wstride,
		 const double *y, const size_t ystride,
		 const size_t n,
		 double *c0, double *c1,
		 double *cov_00, double *cov_01, double *cov_11,
		 double *chisq)
{

  /* compute the weighted means and weighted deviations from the means */

  /* wm denotes a "weighted mean", wm(f) = (sum_i w_i f_i) / (sum_i w_i) */

  double W = 0, wm_x = 0, wm_y = 0, wm_dx2 = 0, wm_dxdy = 0;

  size_t i;

  for (i = 0; i < n; i++)
    {
      const double wi = w[i * wstride];

      if (wi > 0)
	{
	  W += wi;
	  wm_x += (x[i * xstride] - wm_x) * (wi / W);
	  wm_y += (y[i * ystride] - wm_y) * (wi / W);
	}
    }

  W = 0;			/* reset the total weight */

  for (i = 0; i < n; i++)
    {
      const double wi = w[i * wstride];

      if (wi > 0)
	{
	  const double dx = x[i * xstride] - wm_x;
	  const double dy = y[i * ystride] - wm_y;

	  W += wi;
	  wm_dx2 += (dx * dx - wm_dx2) * (wi / W);
	  wm_dxdy += (dx * dy - wm_dxdy) * (wi / W);
	}
    }

  /* In terms of y = a + b x */

  {
    double d2 = 0;
    double b = wm_dxdy / wm_dx2;
    double a = wm_y - wm_x * b;

    *c0 = a;
    *c1 = b;

    *cov_00 = (1 / W) * (1 + wm_x * wm_x / wm_dx2);
    *cov_11 = 1 / (W * wm_dx2);

    *cov_01 = -wm_x / (W * wm_dx2);

    /* Compute chi^2 = \sum w_i (y_i - (a + b * x_i))^2 */

    for (i = 0; i < n; i++)
      {
	const double wi = w[i * wstride];

	if (wi > 0)
	  {
	    const double dx = x[i * xstride] - wm_x;
	    const double dy = y[i * ystride] - wm_y;
	    const double d = dy - b * dx;
	    d2 += wi * d * d;
	  }
      }

    *chisq = d2;
  }

  return GSL_SUCCESS;
}



int
gsl_fit_linear_est (const double x,
		    const double c0, const double c1,
		    const double c00, const double c01, const double c11,
		    double *y, double *y_err)
{
  *y = c0 + c1 * x;
  *y_err = sqrt (c00 + x * (2 * c01 + c11 * x));
  return GSL_SUCCESS;
}


int
gsl_fit_mul (const double *x, const size_t xstride,
	     const double *y, const size_t ystride,
	     const size_t n, 
             double *c1, double *cov_11, double *sumsq)
{
  double m_x = 0, m_y = 0, m_dx2 = 0, m_dxdy = 0;

  size_t i;

  for (i = 0; i < n; i++)
    {
      m_x += (x[i * xstride] - m_x) / (i + 1.0);
      m_y += (y[i * ystride] - m_y) / (i + 1.0);
    }

  for (i = 0; i < n; i++)
    {
      const double dx = x[i * xstride] - m_x;
      const double dy = y[i * ystride] - m_y;

      m_dx2 += (dx * dx - m_dx2) / (i + 1.0);
      m_dxdy += (dx * dy - m_dxdy) / (i + 1.0);
    }

  /* In terms of y =  b x */

  {
    double s2 = 0, d2 = 0;
    double b = (m_x * m_y + m_dxdy) / (m_x * m_x + m_dx2);

    *c1 = b;

    /* Compute chi^2 = \sum (y_i -  b * x_i)^2 */

    for (i = 0; i < n; i++)
      {
	const double dx = x[i * xstride] - m_x;
	const double dy = y[i * ystride] - m_y;
	const double d = (m_y - b * m_x) + dy - b * dx;
	d2 += d * d;
      }

    s2 = d2 / (n - 1.0);	/* chisq per degree of freedom */

    *cov_11 = s2 * 1.0 / (n * (m_x * m_x + m_dx2));

    *sumsq = d2;
  }

  return GSL_SUCCESS;
}


int
gsl_fit_wmul (const double *x, const size_t xstride,
	      const double *w, const size_t wstride,
	      const double *y, const size_t ystride,
	      const size_t n, 
              double *c1, double *cov_11, double *chisq)
{

  /* compute the weighted means and weighted deviations from the means */

  /* wm denotes a "weighted mean", wm(f) = (sum_i w_i f_i) / (sum_i w_i) */

  double W = 0, wm_x = 0, wm_y = 0, wm_dx2 = 0, wm_dxdy = 0;

  size_t i;

  for (i = 0; i < n; i++)
    {
      const double wi = w[i * wstride];

      if (wi > 0)
	{
	  W += wi;
	  wm_x += (x[i * xstride] - wm_x) * (wi / W);
	  wm_y += (y[i * ystride] - wm_y) * (wi / W);
	}
    }

  W = 0;			/* reset the total weight */

  for (i = 0; i < n; i++)
    {
      const double wi = w[i * wstride];

      if (wi > 0)
	{
	  const double dx = x[i * xstride] - wm_x;
	  const double dy = y[i * ystride] - wm_y;

	  W += wi;
	  wm_dx2 += (dx * dx - wm_dx2) * (wi / W);
	  wm_dxdy += (dx * dy - wm_dxdy) * (wi / W);
	}
    }

  /* In terms of y = b x */

  {
    double d2 = 0;
    double b = (wm_x * wm_y + wm_dxdy) / (wm_x * wm_x + wm_dx2);

    *c1 = b;

    *cov_11 = 1 / (W * (wm_x * wm_x + wm_dx2));

    /* Compute chi^2 = \sum w_i (y_i - b * x_i)^2 */

    for (i = 0; i < n; i++)
      {
	const double wi = w[i * wstride];

	if (wi > 0)
	  {
	    const double dx = x[i * xstride] - wm_x;
	    const double dy = y[i * ystride] - wm_y;
	    const double d = (wm_y - b * wm_x) + (dy - b * dx);
	    d2 += wi * d * d;
	  }
      }

    *chisq = d2;
  }

  return GSL_SUCCESS;
}

int
gsl_fit_mul_est (const double x, 
                 const double c1, const double c11, 
                 double *y, double *y_err)
{
  *y = c1 * x;
  *y_err = sqrt (c11) * fabs (x);
  return GSL_SUCCESS;
}
